! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
module closserrbars_mod
!  VNY 02.11.2004 : a function for rigorous computation
!  of the error bars for the coincidence-loss corrected count rates
!  is added:
!
!  cLossErrBars(frameTime, timeDel, observedCountRate, verbFlag, deadFrac)
!
! The function returnes the rms-error estimated with the use of the rigourous algorithm
! for error-bars (S.Rosen, 2004). 
!
! Input parameters:
! ----------------
! frameTime - real(kind=double): frame time in seconds
! timeDel - real: time of the observation, in seconds (it could be a bin-time or the e
! xposure time)
! deadFrac - real(kind=double): dead fraction in units of the fraction of the frame time
! observedCountRate - real : observed count rate (which is affecetd by the coincidence losses)
!
! Output parameters:
!-------------------
! trueCountRate - real: the estimated count rate (corrected for the coincidence losses); 
! this value can be used for checking purposes only
!--------------------------------------------------------------------
! 2013-11-22 vny: adding the function ComputeSimonErrors (from ommag)
!--------------------------------------------------------------------


   use dal
   use cal
   use param
   use errorHandling
   use frameTime_mod
   use datafilter
   use sasdal
   use omconstants_mod
   use omfunctions_mod
   use brent_mod
   use splinefit_mod
   use omphotometrydefs_mod  
   use omcoinccorr_mod

   
   use omsourceparam_mod ! aded 2013-11-22 vny

   implicit none

   private initRandomSequence

   public coincLossAsymErr
   public cLossErrBars

   public computesimonerrors  ! aded 2013-11-22 vny

   real(single), private, pointer :: clossCountRate(:) => null()

   real(single), private, pointer :: clossErrorBar(:) => null()
   type(OneDSpline), private, save ::  cLossErrBarsSpline
   logical :: initialisedCLossErrBars = .false.

   integer(int16), private, parameter :: maxSize = 100  ! aded 2013-11-22 vny
   real(double), private :: cLossFrameTime, cLossDeadFrac ! aded 2013-11-22 vny
   real(single), private :: cLossExposure

interface  cLossErrBars
    module procedure cLossErrBars1, cLossErrBars2
end interface  cLossErrBars


 


real(double), private :: cLossTrueCountRate, cLossFrameTime, emp_corr=0.0_double
real(double), private, save :: ec(5) = (/1.0d0, 0.0756730d0, -0.1439860d0, 0.6384970d0, -0.5701840d0/)

 
contains

!***********************************************************
!
!
!**********************************************************
! 2013-11-21 the routine explicitely using the formula (7) 
! from Simon&Paul's 2008 paper
  subroutine ComputeSimonErrors(nsources,                        & 
       corrCountRate,                   &
       corrCountRateErr,                &
       corrBackgroundCountRate,         & 
       exposure,                        &
       deadFrac,                        &
       frameTime)
    integer(int32) :: nsources
    real(single), pointer, dimension(:) :: corrCountRate, corrCountRateErr
    real(single), pointer, dimension(:) :: corrBackgroundCountRate
    real(single), intent(in) :: exposure
       
    real(double), intent(in) :: deadFrac, frameTime
    real(single) :: corrBackgroundCountRateErr

    integer(int16) :: i
    real(single) :: totRate, totRateErr

    real(single) :: cminRate
    

    real :: rmsArraySimple(2000)


    integer :: maxCounts, nAux, iloss
    real(kind=double) :: maxCountsPerFrame
    real :: sim_frameTime, sim_timeBin, trueCountRate
    real :: maxLoss, rmsValue
    
 
    real(double) :: x

    real(single) :: nframes

    if (nsources .lt. 1) return

    clossexposure = 1.0
 
    cLossFrameTime = frameTime
    cLossDeadFrac = deadFrac

    !write(*,*) "L577 routine ComputeSimonsErrors ... "

    if (frameTime .gt. 0.0) then
       maxCountsPerFrame=0.99/frameTime
    else
       maxCountsPerFrame=200.0
    endif
    maxCounts=6*maxCountsPerFrame
    !write(*,*) "L596 maxCounts=", maxCounts
    if (maxCounts .gt. 2000) maxCounts=2000
    !write(*,*) "L588 nsources=", nsources, " maxSize=", maxSize, " maxCounts=", maxCounts

    sim_timeBin= exposure
    sim_frameTime=frameTime

    nframes=sim_timeBin/frameTime
    !write(*,*) "L609 nframes=", nframes
    
    do i=1,maxCounts
       trueCountRate=i
       
       x=trueCountRate
              
       call ComputeRmsValueSimple(frameTime, deadFrac, sim_timeBin, x, rmsValue)
       rmsArraySimple(i)=rmsValue
       
    enddo


    maxLoss=0.0
    !write(*,*) "L612 =====================================================+ maxCounts=", maxCounts
    do i=1,maxCounts
       trueCountRate=i       
       !write(*,*) "L615 i=", i, " rmsSimple=",rmsArraySimple(i) 
    enddo ! i
    !write(*,*) "L612 =====================================================+ maxCounts=", maxCounts
 

    nAux=int(maxLoss)

    cminRate = 5.0

    !write(*,*) "L702 nsources=", nsources, " (larger than maxSize)"
    cminRate = max(cminRate, minval(corrCountRate + corrBackgroundCountRate, &
         (corrCountRate+corrBackgroundCountRate) .gt. 0.0))
    
    !write(*,*) "L707 cminRate=", cminRate
    
    
    
    do i = 0, nsources - 1
       totRate = corrCountRate(i) + corrBackgroundCountRate(i)
       !write(*,*) ".......... totRate=", totRate
       
       if (totRate .lt. cminRate) then
          
          totRateErr=corrCountRateErr(i)
          !write(*,*) "L129 i=", i, " corrCountRate=", corrCountRate(i), &
          !     " corrBackgroundCountRate=", corrBackgroundCountRate(i), & 
          !     " totRate=", totRate, " totRateErr=", totRateErr, "  ... skipped "
       else
          !***********************************************************
          ! The error is picked up from the initiated array of eerors 
          !***********************************************************         
          iloss=int(totRate+0.55)             
          if (iloss .gt. maxCounts) iLoss=maxCounts
          if (iloss .gt. 0) then
             
             totRateErr=rmsArraySimple(iloss)
          endif
          
          !write(*,*) "L696 i=", i, " corrCountRate=", corrCountRate(i), &
          !     " corrBackgroundCountRate=", corrBackgroundCountRate(i), & 
          !     " totRate=", totRate, " totRateErr=", totRateErr
          
          !*******************************************************
          ! The error in the background  count-rate
          !*******************************************************
          corrBackgroundCountRateErr = sqrt(corrBackgroundCountRate(i)) * sqrt(exposure) / exposure
          corrCountRateErr(i) = sqrt(totRateErr * totRateErr + &
               2.0 * corrBackgroundCountRateErr * corrBackgroundCountRateErr)
          
          !write(*, *) "L707    totRate=", totRate, " corrBkgErr=", corrBackgroundCountRateErr, &
          !     " corrCountRateErr=",   corrCountRateErr(i)          
       endif
    end do
    !write(*,*) "L711 ================================"

end subroutine ComputeSimonErrors

!---------------------------------------------------------------------------------
! Simplified the routine using Simon's coincidence loss test routine
! 2013-11-21 vny
! The simplified algorithm for calculating the CL-corrected count rate
! errors is based on the formula (7) from Kuin&Rosen paper MNRAS 383, 383 (2008).
! It calculates the error based on the true count rate which should be 
! estimated beforehand from the observed count rates by using CAL CL-routine.
! The formula is slightly transformed, so the formula used here is:
! sigma_counts = N_frames * ln{1+sqrt[(exp(trueCountsPerFrame -1)/N_frames]}/(1-deadFrac)
! sigma_countrate=sigma_counts/exposure
!---------------------------------------------------------------------------------
subroutine ComputeRmsValueSimple(frameTime, deadFrac, exposure, trueCountrate, rmsValue)
    real(double), intent(in) :: frameTime ! frame time 
    real(double), intent(in) :: deadFrac ! dead-time fraction
    real(single), intent(in) :: exposure ! total exposure
    real(double), intent(in) :: truecountrate ! [c/s] incident count rate (CL-corrected count rate)
  
    real(single), intent(out) :: rmsValue
 
    real(double) :: nFrames
    real(double) :: rms
    real(double) :: alpha

    real(double) :: trueCountsPerFrame

    rms = 0.0_double
    if (frameTime .gt. 0.0_double) then
       nFrames = dble(exposure) / frameTime
    else
       ! assume the frame time to be the standard 11 ms
       nFrames =  dble(exposure) /0.011_double
    endif

    ! dead-time correction factor:
    alpha=1.0_double-deadFrac
    if (alpha .lt. 0.5_double) then
       alpha=1.0_double
    endif

    trueCountsPerFrame=trueCountrate*frameTime
    if (nFrames .gt. 1.0_double) then       
       rms=1.0_double+sqrt((dexp(trueCountsPerFrame)-1.0_double)/nFrames)
    else
       rms=1.0_double+sqrt(dexp(trueCountsPerFrame)-1.0_double)
    endif
    rms=dlog(rms)*nFrames/alpha

    if (exposure .gt. 1.0) then
       rmsValue=real(rms)/exposure
    else
       rmsValue=real(rms)
    endif 
end subroutine computeRmsValueSimple

!*************************


!******************************************************************
!
! Coincidence-loss emperical correction function
!******************************************************************
real(double) Function clecFunction(muo)
    real(double), intent(in) :: muo
    
    real(double) :: ci_pred
    real(double) :: muoSquared

    muoSquared = muo * muo
   

    emp_corr = ec(1) + muo * (ec(1) + ec(2) * muo) + muoSquared * muo * (ec(3)  + & 
               ec(4) * muo +  ec(5) * muoSquared)


    ci_pred = (-log(1.d0 - muo) / cLossFrameTime) * emp_corr
  
    clecFunction = abs(ci_pred - cLossTrueCountRate) 


end function clecFunction




!----------------------------------------------------------------------------
! Using Simon's algorithm for CL-correction and estimation of rms
!----------------------------------------------------------------------------
! The function returnes the rms-error for the estimated CL-corrected counts 
! 
!
! Input parameters:
! ----------------
! frameTime - real(kind=double): frame time in seconds
! timeDel - real: time of the observation, in seconds (it could be a bin-time or 
! the exposure time)
! deadFrac - real(kind=double): dead fraction in units of the fraction of the frame time
! observedCountRate - real : observed count rate (which is affecetd by the 
! coincidence losses)
!
! Output parameters:
!-------------------
! trueCountRate - real: the estimated count rate (corrected for the coincidence losses); 
! this value can be used for checking purporses only
!
!
! NB cminRate and cMaxRate are raw count-rates
!**********************************************************************************
subroutine InitialiseCLossErrBars(exposure, cminRate, cmaxrate)

    
    real, intent(in) :: exposure
    real(single), intent(in) :: cminRate, cmaxRate

    integer :: i
    real(single) :: dr
    integer, parameter :: maxSize = 25
    real(single) :: rawCountRate, sqrtExposure

    if (associated(clossCountRate)) return
    write(errstr,'(a)') "Initialising CLossErrBars"    
    call message(AppMsg, Verbose)
    exposureLength = exposure
    sqrtExposure = sqrt(exposure)
    dr = (cMaxRate - cminRate) / real(maxSize - 1)


    if (.not. associated(clossErrorBar)) then
        clossCountRate=> allocateSingle(1, maxSize, "clossCountRate")

        clossErrorBar => allocateSingle(1, maxSize, "clossErrorBar")

    end if
write(*, *) "closserrbars"
    do i = 1, maxSize
      
        rawCountRate =  cMinRate + real(i - 1) * dr

        ! Compute the error for this count-rate
        cLossErrorBar(i) = cLossErrBars(rawCountRate / sqrtexposure, frameTime, 1.0, &
                           cLossCountRate(i))  ! NB clossCountRate stoes the corrected count-rate
write(*, *) i, " corr rate=", clossCountRate(i), " error=", clossErrorBar(i), " sqrt=", sqrt(clossCountRate(i))     
        cLossCountRate(i) =  cLossCountRate(i) * sqrtexposure
        clossErrorBar(i) = clossErrorBar(i) * sqrtexposure
    end do

    ! Construct spline interpolant
    call constructSpline(maxSize, cLossCountRate, cLossErrorBar, cLossErrBarsSpline)

    initialisedCLosserrBars = .true.  

end subroutine InitialiseCLossErrBars
!*********************************************************
!
!
!********************************************************
real(single) function cLossErrBars1(countRate)
    real(single), intent(in) :: countrate
  

    if (initialisedCLossErrBars) then
        clossErrBars1 = interpolateSpline(countRate, cLossErrBarsSpline)
    else

        cLossErrBars1 = sqrt(countrate)
               
    end if


end function cLossErrBars1

!*********************************************************
!
!
!********************************************************

real(single) function cLossErrBars2(rawCountRate, frameTime, timeDel, corrRate)
    real(double), intent(in) :: frameTime ! Seconds
    real(single), intent(in) :: timeDel
 
    real(single), intent(in) :: rawCountRate
    real(single), intent(out) :: corrRate
    real(single) :: dispersionLeft
    real(single) :: dispersionRight
    real(single) :: rmsValue, rmsValue1, rmsValue2
    real(double) :: maxCountsPerFrame, sourceErr, corrRateErrMinus, corrRateErrPlus
    integer :: maxCounts, i, minCounts
    real :: sim_frameTime, sim_timeBin
    real :: lossCountRate1, lossCountRate2
    real(single) :: deltaLoss, truecountRate
    real(single) :: countRateError
    real(double) :: countRateErrorMinus, countrateErrorPlus


    ! Correct for coincidence-loss  
    countRateError = sqrt(rawCountRate)

    corrRate = CorrectForCoincidenceLoss(dble(rawCountRate), &
                  countRateError, frameTime,  &
                  countRateErrorMinus, countRateErrorPlus)


    if (.not. associated(factorialData%computed)) then
        call InitialiseFactorial
    end if


    if (frameTime .gt. 0.0) then
        maxCountsPerFrame = 0.99 / frameTime
    else
       maxCountsPerFrame = 200.0
    end if

    maxCounts = 6.0 * maxCountsPerFrame

    maxCounts = min(2000, maxCounts)

    sourceErr = 0.0
    corrRateErrMinus = 0.0
    corrRateErrPlus = 0.0
   
     
    sim_timeBin = timeDel
    sim_frameTime = frameTime



    ! Check the values of corrRate from corrRate-10% to maxCounts
    minCounts = corrRate - 0.1 * corrRate

    do i = minCounts, maxCounts - 1

        trueCountRate = real(i)
        call coincLossAsymErr(sim_frameTime, sim_timeBin, trueCountRate, &
              lossCountRate1, dispersionLeft, &
              dispersionRight, rmsValue1)

        trueCountRate = real(i + 1)
        call coincLossAsymErr(sim_frameTime, sim_timeBin, trueCountRate, &
              lossCountRate2, dispersionLeft, &
              dispersionRight, rmsValue2)
        if (lossCountRate1 .le. corrRate .and. &
              lossCountRate2 .gt. corrRate) exit  
    end do

    deltaLoss = lossCountRate2 - lossCountRate1

    if (abs(deltaLoss) .gt. 1.0e-5) then
        rmsValue = rmsValue1 + (rmsValue2 - rmsValue1) * (lossCountRAte2 - rawCountRate) / deltaLoss
    else 
        rmsValue=rmsValue1
    endif

    cLossErrBars2=rmsValue
end function cLossErrBars2


!---------------------------------------
! Using Simon's coincidence loss test routine
! 27.05.2004
!---------------------------------------
subroutine coincLossAsymErr(frameTime, timeBin, trueCountRate, &
     lossCountRate, dispersionLeft, dispersionRight, rmsValue)
    real, intent(in) :: frameTime
    real, intent(in) :: timeBin
    real, intent(in) :: trueCountRate ! counts per second
    real, intent(out) :: lossCountRate
    real, intent(out) :: dispersionLeft
    real, intent(out) :: dispersionRight
    real, intent(out) :: rmsValue
 
    integer :: nFrames


    real(kind=double) :: mu, mu_eff, ln_p, p
    real(kind=double) :: ln_k_fact, ln_m_fact, ln_k_m_fact, sum_p
    real(kind=double) :: mu_poiss
    real(kind=double) :: rms, mean, r2, errl, errh, devl, devh
    real(kind=double) :: cl_formula, sum_p_l, sum_p_h
    real(kind=double) :: ln_fixed_term, crfrr
    real(kind=double) :: fco

    integer(kind=int32) :: k, i

    real(kind=double) :: ci, t ! incident CR (cts/s) and frame rate (s)
    real(kind=double) :: co_best
    real(kind=double) :: ci_pred, muo, mubest
    real(double) :: logoneminusexpminusmu
    real(double) :: f
    real(double) :: iMinusMean


! VNY 14.06.2004 from Simon's code:
! insert empirical coincidence loss coefficients, from 
! ~/sasccf/OM_PHOTTONAT_0003.CCF. Coefficients work on the co*t (raw CRFR 
! ratio).



! write(*, *) "coincLossAsymErr line 443"


    nFrames = int(timeBin / frameTime)

    ci = trueCountRate

    t = frameTime

    mu_poiss = ci ! incident counts during one second
    mu= ci * t ! incident counts during one frame
    crfrr = ci * t ! INCIDENT CRFR ratio
    cl_formula = (1.d0 - exp(-mu)) / frameTime !t

    k = nFrames ! total number of frames per exposure time (timeBin)


! There is a PROBLEM to use the empirical curve directly here in this 
! test code because the empirical correction uses a polynomial based on 
! the observed, not incident count rate so have to iteratively find what 
! observed CR gives incident count rate provided by user. Quick solution 
! here where we have ci, not co, initially, is to simply compute ci for 
! a range of co*T (=muo =CRFRobserved values) and find value of co which 
! gives Ci (empirical) closest to input value



    !***************************************************
    ! FOR TEST
    !*************************************************
    cLossframeTime = t

    clossTrueCountRate = trueCountRate 

    f = brent(0.0_double, 1.0_double, 0.0001_double,  mubest, clecFunction)


    co_best = mubest / t

    ci_pred = -log(1.d0 - mubest) / t
    ci_pred = ci_pred * emp_corr
     

! compute correction factor f(Co) that is thus used to convert this Co to Ci
! Will be used lower down if empirical correction is applied. 

    muo = co_best * t
    fco = ec(1) + ec(2)*muo + ec(2)*muo*muo + ec(3)*muo**3 + & 
          ec(4)*muo**4 + ec(5)*muo**5
      
! First compute the probability distribution of the incident count rate, for 
! reference
!

! As a check, compute sigma using sqrt[sum(p*(xi-X)**2] rather than 
! sqrt[bar(r**2) - (bar(r))**2]
!
    devl = 0.d0
    devh = 0.d0
    sum_p_l = 0.d0
    sum_p_h = 0.d0


!  compute the probability of getting m=i counts if we run k frames with an 
!  incident rate of Ci counts/s and a frame time of t seconds. This 
!  probability is based on kCm * P(0)**(k-m) * P(>=1)**m 
!
!  where
!
!  kCm = k!/(m!)(k-m)!
!
!

    sum_p = 0.d0
    rms = 0.d0
    r2 = 0.d0
    mean = 0.d0

    mu_eff = mu

    mu_eff = mu / fco   !compute effective incident mu
    cl_formula = co_best
 
    ln_fixed_term = log(1.d0 - exp(-mu_eff)) !this term if fixed for each k below



    ln_k_fact = logFactorial(k)     ! constant for given no of frames
!write(*, *)   "ln_k_fact =",   ln_k_fact 
 !write(*, *) "coincLossAsymErr line 557"

    do i = 0, k  ! m can not exceed k
             
        ln_m_fact = logFactorial(i)

        ln_k_m_fact = logFactorial(k - i)


        ln_p = ln_k_fact - ln_m_fact - ln_k_m_fact

        ln_p = ln_p - (k - i) * mu_eff + i * ln_fixed_term 
        p = exp(ln_p)
        if(p .lt. 1.d-30) p = 0.d0    ! set to zero if P is very small. 
        mean = mean + p * i
        r2 = r2 + p * i * i
        sum_p = sum_p + p
!write(*, *) "i=", i, " ln_m_fact=",  ln_m_fact, "   ln_k_m_fact=",  ln_k_m_fact 
    end do

!call fatal("stopped")
!--- 15.06.2004
! rounding errors and Stirlings approx give rise to value of sum_p which is 
! very slightly > 1.000. Thus, normalise sum(r2) amd mean by sum_p as if to 
! make them equivalent to having come from a distribution with sum_p=1.0. 
! This appears to remove potential problem of sum(r2)-mean*mean being -ve. 

    r2 = r2 / sum_p         
    mean = mean / sum_p     
    rms = r2 - mean * mean    ! rms (square of)
    

 !write(*, *) "coincLossAsymErr line 588, sum_p=", sum_p

!
! now compute asymmetric errors - For asymmetric errors need to know 
! what mean is first, so must compute above loop first and then repeat to 
! get deviations from mean, weighted by P
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    errl = 0.d0
    errh = 0.d0
    devl = 0.d0
    devh = 0.d0
    rms = 0.d0

    logoneminusexpminusmu = log(1.0d0 - exp(-mu))

!write(*, *) "Line 646 mean=", mean, " sum_p=", sum_p
    do i = 0, k     ! i can not exceed k
    
        ln_m_fact = logFactorial(i)

        ln_k_m_fact = logFactorial(k - i)
        ln_p = ln_k_fact - ln_m_fact - ln_k_m_fact
        ln_p = ln_p - dble(k - i) * mu + dble(i) * logoneminusexpminusmu
        p = exp(ln_p)
        if (p .lt. 1.d-30) p = 0.d0
        iMinusMean = dble(i) - mean
        iMinusMean = iMinusMean * iminusMean
        
        rms = rms + p * iMinusMean 

        if (i .le. mean) then
            devl = devl + p * iMinusMean
        else
            devh = devh + p * iMinusMean
        end if
    end do

    errl = sqrt(devl)
    errh = sqrt(devh)
 
    if (rms .gt. 0.d0) then
        rms = sqrt(rms)
  
    endif

    rmsValue = rms / timeBin
    dispersionLeft = errl / timeBin
    dispersionRight = errh / timeBin
    lossCountRate = mean / timeBin

! write(*, *) "coincLossAsymErr line 683 timebin=", timebin, " errl=", errl, " errh=", errh


end subroutine coincLossAsymErr

! End of Simon's code
!*****************************************************************

!--------------------------------------
! Initialise the random number generator
!--------------------------------------
subroutine initRandomSequence(initNumber)
  integer, intent(in) :: initNumber
  integer :: seedSize, i
  integer, allocatable :: seedArray(:)
  call random_seed(SIZE=seedSize)
  allocate(seedArray(seedSize))

  do i=1,seedSize
     seedArray(i)=i*initNumber
  enddo
  call random_seed(PUT=seedArray)
  deallocate(seedArray)
end subroutine initRandomSequence

end module closserrbars_mod

