! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
module omcelestial_mod

  use oal
  use cal
  use dal
  use sasdal
  use caloaldefs

  use errorHandling
  use sla

 
  use omconstants_mod
  use omdistortion_mod
contains


!*******************************************************************
!
!
!******************************************************************
subroutine ComputeCelestialCoordinates(xPos, yPos, x0, y0, &
                                       xBinning, yBinning, &
                                       plateScale, rollAngle, &  
                                       ra_point, dec_point, &
                                       rightAscension, declination , &
                                       galacticLong, galacticLat)
    real(single), intent(in) :: xPos, yPos
    real(single), intent(in) :: x0, y0
 
    real(single), intent(in) :: xBinning, yBinning
    real(double), intent(in) :: plateScale, rollAngle
    real(double), intent(in) :: ra_point, dec_point
  
    real(double), intent(out) :: rightAscension, declination
    real(double), intent(out) :: galacticLong, galacticLat
    real(double) ::  xposcorr,  yposcorr
    real(double) :: xi_source, eta_source
    real(double) :: xrad, yrad              ! position in radians     
    real(double) :: cosRollAngle, sinRollAngle
    integer(int32) :: mode

    cosRollAngle = dcos(rollAngle)
    sinRollAngle = dsin(rollAngle)
      

    xposcorr = dble(xpos) * dble(xBinning) + dble(x0)  
    yposcorr = dble(ypos) * dble(ybinning) + dble(y0)
 
    !***************************************************
    ! correct positions for pincushion/barrel distortion
    !***************************************************
  
    rawx(0) = real(xposcorr)
    rawy(0) = real(yposcorr)

    !write(*,*)'Undistort: ', xposcorr,yposcorr

    mode = 1
    call CAL_omDistortion(rawx, rawy, deltax, deltay, sigmax, sigmay, mode)


    xposcorr = xposcorr - dble(deltax(0))  ! - sign removes distortion    
    yposcorr = yposcorr - dble(deltay(0))  ! - sign removes distortion

    !write(*,*)'XPOS: ', xposcorr,yposcorr
           
    !***************************************************
    ! get offset relative to tangent point (at OM centre)
    !***************************************************
    xrad = xposcorr - xCentre 
    yrad = yposcorr - yCentre

    !write(*,*)'ROT: ', xrad, yrad

    !**************************************************
    !  rotate x,y frame to tangent sky (RA, DEC) frame
    !  and form tangent-plane coords
    !*************************************************
    xi_source = -plateScale * (xrad * cosRollAngle - yrad * sinRollAngle)
    eta_source = plateScale * (xrad * sinRollAngle + yrad * cosRollAngle)


    !***********************************************************
    ! Transform tangent-plane coordinates (xi_source, yi_source) 
    ! into spherical (sky) coordinates (RA, dec). ra_point,
    ! dec_point is the spherical coordinate of the tangent-point
    ! Also calculate the galactic longitude and latitude of
    ! each source
    !************************************************************
    !write(*,*)'Into p2s: ', xi_source, eta_source
    call sla_dtp2s(xi_source, eta_source, &
            ra_point, dec_point, rightAscension, declination)

    call sla_EQGAL(rightAscension,  declination, &
                      galacticLong, galacticLat)

end subroutine ComputeCelestialCoordinates

!*******************************************************************
! ComputeXYCoordinates - converts from RA/DEC to pixel coords
!                      - the inverse of ComputeCelestialCoordinates
!******************************************************************
subroutine ComputeXYCoordinates(rightAscension, declination, &
                                       ra_point, dec_point, &
                                       x0, y0, &
                                       xBinning, yBinning, &
                                       plateScale, rollAngle, &
                                       xPos, yPos)
    real(double), intent(in) :: rightAscension, declination !radians
    real(double), intent(in) :: ra_point, dec_point !radians
    integer(int32), intent(in) :: x0, y0 ! WINDOWX0, WINDOWY0 keywords
    integer(int32), intent(in) :: xBinning, yBinning ! 2**BINAX1, 2**BINAX2
                                                   ! *2 (if BINBPE=T)
    real(double), intent(in) :: plateScale  ! Units of arcsecs/pix ?
    real(double), intent(in) :: rollAngle ! Units of radians
    real(double), intent(out) :: xPos, yPos
    !real(double) :: xPos, yPos
    real(double) :: xrad,yrad
    integer(int32) :: istat 
    real(double) :: xunrot, yunrot
    real(double) :: cosRollAngle, sinRollAngle
    integer(int32) :: mode

    cosRollAngle = dcos(rollAngle)
    sinRollAngle = dsin(rollAngle)

    !convert from RA,DEC to tangential coords
    call sla_ds2tp(rightAscension, declination, &
                       ra_point, dec_point, xrad, yrad, istat)
    !write(*,*)'SLA: ',rightAscension, declination,xrad,yrad,istat

    !! At this stage - ommosaic/convert4 does:
    !x = x0 + dx / plateScale  ! x0 is in pixels, platescale in radians 
    !y = y0 - dy / plateScale  ! y0 is in pixels, platescale in radians
    ! and thinks that it has finished

    !unrotate frame to x,y frame
    !xi_source = -plateScale * (xrad * cosRollAngle - yrad * sinRollAngle)
    !eta_source = plateScale * (xrad * sinRollAngle + yrad * cosRollAngle)
    xunrot = 1.0/plateScale * (-cosRollAngle*xrad + sinRollAngle * yrad)
    yunrot = 1.0/plateScale * (xrad * sinRollAngle + yrad * cosRollAngle)

    !write(*,*)'Unrot: ',xunrot,yunrot,xCentre,yCentre
    ! Add on the centre value
    xunrot = xunrot + xCentre
    yunrot = yunrot + yCentre

    !if successful then apply the distortion map to get detector position
    if (istat .eq. 0) then
    !rawx(0) = real(xunrot)
    !rawy(0) = real(yunrot)
    !mode=1
       call InitialiseDistortion(0,0,1,1) ! Input already in FFU coords
        !call CAL_omInverseDistortion(rawx,rawy,deltax,deltay,&
         !    sigmax, sigmay,mode)
       !write(*,*)'OR: ',xunrot+deltax,yunrot+deltay
       call AddDistortion(xunrot,yunrot, xPos, yPos)
       !write(*,*)'Distortion: ',xunrot,yunrot,xPos, yPos
    else
       call fatal("Error calling sla_ds2tp!")
       ! with RA=", rightAscension," DEC=",declination)
    endif

end subroutine ComputeXYCoordinates

end module omcelestial_mod



