! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!**************************************************************
! module omdetectcoinccorr_mod.f90 (15th August 2002)
!
! Contains routines for correcting count rates
! for coincidence losses
!
!**************************************************************
! 2013-12-04 vny: applying the limit of 0.97 counts/frame to the 
! count rates. Above this limit the coincidence-loss corrected
! counts are not calibrated and have very large errors
!**************************************************************
module omcoinccorr_mod
    
    use param
    use cal
    use dal
    use omsourceparam_mod
    use omutility_mod



!*************************************************
!
!
!
!************************************************
real(single), private :: significanceRadius = 6.0


!****************************************************
!
!
!
!****************************************************
interface SetSignificanceRadius
    module procedure  SetSignificanceRadius1, SetSignificanceRadius2
end interface SetSignificanceRadius
   

contains               
!******************************************************
!
!
!******************************************************

real(single) function getSignificanceRadius()

    getSignificanceRadius = significanceRadius

end  function getSignificanceRadius


!******************************************************
!
!
!******************************************************
subroutine setSignificanceRadius1( radius )
    real(single), intent(in) :: radius
    significanceRadius = radius

end  subroutine setSignificanceRadius1


!***************************************************
! SetSignificanceRadius
! Set the radius for which the source-significance
! is computed (signal-to-noise)
! The radii are hard-wired so that for "V" the
! radius is 6 unbinned pixels, and for the other
! OM filters the radius is such that the fraction
! of the total counts contained within that radius
! is the same as the V-band value (95.4%)
! 
!
!***************************************************
subroutine setSignificanceRadius2(filter)
    character(*), intent(in) :: filter


    select case(trim(filter))
    
        case ("U")
            significanceRadius = 6.48
        case ("B")
            significanceRadius = 6.66

        case ("V")
            significanceRadius = 6.00

        case ("UVW2")
            significanceRadius = 8.38

        case ("UVM2")
            significanceRadius = 8.41
        case ("UVW1")
            significanceRadius = 8.72

        case default
            significanceRadius = 6.0

    end select



end subroutine setSignificanceRadius2

!*********************************************************
! Function CoincidenceLossCorrection
!   
!*********************************************************
real(double) function CoincidenceLossCorrection(rawCountRate, & 
                  countRateError, frameTime)
    real(double), intent(in)     :: rawCountRate
    real(single), intent(in)     :: countRateError
    real(double), intent(in)     :: frameTime
    real(single) :: tempRawRate
    
    type(ValueAsymmetricErrorT) :: rawRate
    type(ValueAsymmetricErrorT) :: corrRate
 
    if (rawCountRate .le. 0.0 .or. isNullValue( rawCountRate )) then
        CoincidenceLossCorrection = 0.0_double
        return
    end if




    if( frameTime < 1e-6) then

        CoincidenceLossCorrection = 1.0_double
        return 

    end if


    !*************************************************************
    ! NB frame time is in milli-seconds
    !*************************************************************    
    ! 2013-12-04 setting the max counts per frame to 0.97
    !tempRawRate = min(0.9999_double / frameTime, dble(rawCountRate))
    tempRawRate = min(0.97_double / frameTime, dble(rawCountRate))


!    if (rawCountRate .gt. 1.0_double / frameTime) then
!        CoincidenceLossCorrection = 1.0_double
!        return
!    end if
    
    rawRate = ValueAsymmetricErrorT(temprawRate, dble(countRateError), dble(countRateError))


    !******************************************************************************************
    ! NB the final boolean parameter switches on the empirical correction
    !*****************************************************************************************
    corrRate = CAL_omPhotoNatural(rawRate, frameTime, 0.0_double, 0.0_double, .FALSE., .TRUE.)
 

    !****************************************************************
    ! If the count rate is greater than the frame rate then the above
    ! routine cannot do the correction- instead it sets the corrected 
    ! count rate to the largest number that can be stored on the computer. 
    !*****************************************************************

    !if (CorrRate%value .lt. 1e10) then
    if (CorrRate%value .lt. 1e10 .and. rawCountRate .le. tempRawRate) then
        CoincidenceLossCorrection = corrRate%value / rawRate%value       
    else
        CoincidenceLossCorrection = 1.0_double
    end if
   
    

end function CoincidenceLossCorrection



!****************************************************
! subroutine PointSourceCoincLossCorr
! Corrects a point-source count-rate for coincidence-losses
! 
!****************************************************
! 2013-10-14 passing deadFrac as parameter
!subroutine PointSourceCoincLossCorr(source, coincidenceLossRadius, frameTime)     
subroutine PointSourceCoincLossCorr(source, coincidenceLossRadius, frameTime, deadFrac)     
    type(SourceParameters), intent(inout) :: source

    real(single), intent(in) :: coincidenceLossRadius

    real(double), intent(in) :: frameTime
    real(double), intent(in) :: deadFrac
    real(single) :: countFrameRatio, countsRatio
    real(double) :: totalRate
    real(single) :: error, totRate
    real(double) :: corrTotalRate, corrRateErr, corrBkgRateErr
    real(double) :: corrRateErrMinus1, corrRateErrPlus1
    real(double) :: corrRateErrMinus2, corrRateErrPlus2
    real(double) :: oldCorrRate
    !tttttttttttt
    real(double) :: scalingFactor1, scalingFactor2
    real(double) :: corrRateErrTest
    real(double) :: oldCorrRateErr, new2oldErrRatio
    real(double) :: oldCorrBkgRateErr, new2oldBkgErrRatio
    real(double) :: newCorrRateErr, newCorrBkgRateErr

    real(double) :: new2oldBkgRateErr, new2oldSrcRateErr
    real(double) :: new2oldBkgCorrRate, new2oldBkgCorrRateErr
    real(double) :: new2oldTotRate, new2oldRawRate

    real(double) :: aux

    real(double) :: s2nratio

    ! variables to implement Antonio's foormula:
    !real(double) :: 

!    real(single) :: sourceRate, backgroundRate

!integer, save :: count= 0 

!logical check

    real(double):: deadTimeCorr
    deadTimeCorr=1.0_double/(1.0_double-deadFrac)
 
    countsRatio=1.0

    !write(*,*) "L224 PointSourceCoincLossCorr deadTimeCorr=", deadTimeCorr
    !************************************************************************
    !
    ! Compute the significance (signal-to-noise) using the extracted counts
    !
    !************************************************************************
 
    if(frameTime > 1e-6) then 
        source%significance = computeSourceSignificance(source%rawCountRate, &
             source%rawBackgroundCountRate, source%exposure)
        !write(*,*) "L232 significance=", source%significance
    end if
    !write(*,*) "L234 apertureRadius=", source%apertureRadius, " coincidenceLossRadius=", &
    !     coincidenceLossRadius
    !**********************************************************************
    ! Check to see if the aperture radius was less than the
    ! coincidence-loss radius- apply a correction if so
    !*********************************************************************
    if (source%apertureRadius .lt. coincidenceLossRadius) then

       !write(*,*) "L242 extrapolating counts to CL-radius"
       !*********************************************************
       ! Extrapolate counts from the omdetect aperture radius
       ! to 12-pixel radius and also the background counts
       !**********************************************************
       if(frameTime > 1e-6) then
          countFrameRatio = min(real(source%rawCountRate * frameTime), 1.0)       
          countFrameRatio = max(0.0, countFrameRatio)
       else
          countFrameRatio = 0.1
       end if
       countsRatio = PSFCountsRatio(coincidenceLossRadius, source%apertureRadius, &
            source%xPos, source%yPos, countFrameRatio)
       !write(*,*) "L254 countsRatio=", countsRatio, " corrCountRateErr=", source%corrCountRateErr
       
       source%corrCountRate = source%rawCountRate * countsRatio
       source%corrCountRateErr = source%rawCountRateErr * countsRatio
       !write(*,*) "L258 corrCountRateErr=", source%corrCountRateErr

       !***************************************************************************
       ! Extrapolate the background to the coincidence-loss radius
       !**************************************************************************
       !write(*,*) "L264 rawBackGroundCountRateErr=",source%rawBackGroundCountRateErr

       !*****************************************************
       ! 2013-11-28 calculate signal/noise ratio to be used
       ! in the calculation of the corrected error bars
       !****************************************************
       s2nratio=1.0_double
       if (source%rawCountRateErr>0.0) then
          s2nratio=source%rawCountRate/source%rawCountRateErr
       endif
       if (s2nratio<0.01_double) then
          s2nratio=0.01_double
       endif

       !write(*,*) "L276 source rawCountRate=", source%rawCountRate, " err=", source%rawCountRateErr
       !write(*,*) "L276 s2nratio=", s2nratio

       source%corrBackGroundCountRate = source%rawBackGroundCountRate * &
            (coincidenceLossRadius / source%apertureRadius) ** 2 
       !write(*,*) "L268 rawBackGroundCountRate=",source%rawBackGroundCountRate, &
       !     " corrBackGroundCountRate=",source%corrBackGroundCountRate 

       !ttttttttttttttttttttttttttt 2013-10-02
       scalingFactor1= (coincidenceLossRadius / source%apertureRadius) ** 2
       !write(*,*) "L273 scalingFactor1=", scalingFactor1

       !*********************************************
       ! 2013-11-23 vny: the background rate error will remain almost the same
       ! as before because we are not measuring beckground using the CAL aperture.
       ! The measured aperture remains the same
       !**************************************************************************
       !source%corrBackGroundCountRateErr=source%rawBackGroundCountRateErr * &
       !     (coincidenceLossRadius / source%apertureRadius) ** 2  
       source%corrBackGroundCountRateErr=source%rawBackGroundCountRateErr

       !write(*,*) "L277 corrBackGroundCountRateErr=", source%corrBackGroundCountRateErr
       oldCorrBkgRateErr=source%corrBackGroundCountRateerr
       !ttttttttttttttttttttttttttt 2013-10-02
       totRate = source%corrCountRate + source%corrBackGroundCountRate
       !write(*,*) "L290 totRate=", totRate
       corrRateErrTest=sqrt(source%corrCountRateErr*source%corrCountRateErr+ &
            source%corrBackGroundCountRateerr* source%corrBackGroundCountRateerr)
       !write(*,*) "L291 corrRateErrTest=", corrRateErrTest

       if (totRate .gt. 0.0 ) then         
          
          !*******************************************************************
          ! vny 2013-10-01 Here the calculation of the error seems to be wrong
          ! because the total rate was scaled and was not taken from 12 pixels;
          ! in addition to that, adding the background to totrate is another 
          ! mistake because the background was already added to totRate
          !*******************************************************************
          ! vny 2013-10-23 Restoring the old calculation as the original 
          ! formula corresponds to the correct expression err=sqrt(s+2b)
          ! The extra background value takes into account the background-subtraction
          ! operation which increases the error
          !**********************************************************************
          
          !source%corrCountRateErr = sqrt(totRate + source%corrBackgroundCountRate) &
          !     * sqrt(source%exposure) / source%exposure
       
          !write(*,*) "L302 totRate=", totRate, " corrBackgroundCountRateErr=", source%corrBackGroundCountRateErr
          !*****************************************************************
          ! 2013-11-23 vny: The procedure includes the background addition,
          ! CL-correction, and, finally, the background subtraction.
          ! The background addition and background subtraction cancel eacho other's 
          ! errors, so the formal error propagation should not take into sccount these
          ! addition-subtraction operations.
          !*************************************************************************
         
          !source%corrCountRateErr = sqrt(totRate + source%corrBackGroundCountRateErr* &
          !     source%corrBackGroundCountRateErr) / sqrt(source%exposure)
          
          !write(*,*) "L305 exposure=",source%exposure, " new corrCountRateErr=", source%corrCountRateErr 
          
       else
          !write(*,*) "L346 totRate <0"
          source%corrCountRateErr = source%rawCountRateErr
          source%corrBackgroundCountRateErr = source%rawBackgroundCountRateErr
       end if
       
       !***********************************************************************************
       ! NEED TO CHANGE THIS DEFINITION TO A RATIO FOR CONSISTENCY WITH OTHER CORRECTIONS
       !***********************************************************************************
       source%psf1Correction = (countsRatio - 1.0) * 100.0
    else
       !write(*,*) "L356 using rawCountRate without scaling "
       source%corrCountRate = source%rawCountRate
       source%corrCountRateErr = source%rawCountRateErr
       source%corrBackGroundCountRate = source%rawBackGroundCountRate
       source%corrBackGroundCountRateerr = source%rawBackGroundCountRateerr
       source%psf1Correction = 0.0

       totRate = source%corrCountRate + source%corrBackGroundCountRate
       !write(*,*) "L364 totRate=", totRate, " source corrCountRateErr=", &
       !     source%corrCountRateErr

       !*****************************************************
       ! 2013-12-01 calculate signal/noise ratio to be used
       ! in the calculation of the corrected error bars
       !****************************************************
       s2nratio=1.0_double
       if (source%rawCountRateErr>0.0) then
          s2nratio=source%rawCountRate/source%rawCountRateErr
       endif
       if (s2nratio<0.01_double) then
          s2nratio=0.01_double
       endif
       !write(*,*) "L378 s2nratio=", s2nratio

       !if (totRate .gt. 0.0 ) then         
       !   !****************************************************
       !   ! 2013-11-23 vny: excluding formal error propagation for the background errors
       !   ! because the background addition/subtraction operations cancel the errors
       !   !*************************************************************************
       !   !write(*,*) "L324 totRate=", totRate, " corrBackgroundCountRateErr=", source%corrBackGroundCountRateErr
       !   !source%corrCountRateErr = sqrt(totRate + source%corrBackGroundCountRateErr* &
       !   !     source%corrBackGroundCountRateErr) / sqrt(source%exposure)
       !   !write(*,*) "L327 exposure=",source%exposure, " new corrCountRAteErr=", source%corrCountRateErr 
       !   
       !else
       !   !write(*,*) "L330 totRate <0"
       !   source%corrCountRateErr = source%rawCountRateErr
       !   source%corrBackgroundCountRateErr = source%rawBackgroundCountRateErr
       !end if
             
    end if


    !************************************************************
    ! The total count-rate (source+background)
    !************************************************************
    totalRate = source%corrCountRate + source%corrBackGroundCountRate
    !write(*,*) "L403 totalRate=", totalRate, " source corrCountRateErr=", &
    !     source%corrCountRateErr 


    !***************************************************************
    ! The error in the total rate
    !***************************************************************
    error  = sqrt(totalRate) * sqrt(source%exposure) / source%exposure 
    oldCorrRateErr=  error        
    !write(*,*) "L412 error=", error, " (of the total rate)"

    !-----------------------------------
    ! test error 2013-11-26
    error = sqrt(source%corrCountRateErr*source%corrCountRateErr+ &
         source%corrBackGroundCountRateErr*source%corrBackGroundCountRateErr)
    !write(*,*) "L418 test sqrt(se*se+be*be) error=", error
    !------------------------------------

    !****************************************************************
    ! Correct the total (source + background) for coincidence-loss
    ! NB if the source is "extended" then the total source counts
    ! will have been corrected for coincidence-loss correction by
    ! omdetect, but not the background
    !****************************************************************    
    corrTotalRate = CorrectForCoincidenceLoss(totalRate, error, frameTime, &
                  corrRateErrMinus1, corrRateErrPlus1) 
    !write(*,*) "L429 corrTotalRate=", corrTotalRate, " errMinus=", corrRateErrMinus1, &
    !     " errPlus=", corrRateErrPlus1

    new2oldTotRate=1.0_double
    if (totalRate .gt. 0.0) then
       new2oldTotRate=corrTotalRate / totalRate 
    endif
    !write(*,*) "L436 new2oldTotRate=", new2oldTotRate

    !*****************************************************************
    ! 2013-11-28 vny: the ratio new2oldTotRate, as well as the ratio
    ! new2oldRawRate alters the source S/N ratio. Since this ratio
    ! by definition should remain the same after CL-corrections,
    ! the factor to derive the CL-corrected errors should be based on the
    ! original S/N ratio for raw counts.
    !******************************************************************


    !******************************************************************
    ! Temporarily store correction factors in RA and Dec struct fields
    ! since no entries for them yet
    !******************************************************************

    if (totalRate .gt. 0.0) then
        source%RightAscension= corrTotalRate / totalRate 
        source%coincidenceLossCorr = corrTotalRate / totalRate 
    else
        source%RightAscension = 0.0
        source%coincidenceLossCorr = 0.0 
    end if
    !write(*,*) "L459 sourceRightAscension=", source%RightAscension, &
    !     " coincidenceLossCorr=", source%coincidenceLossCorr  

    !**********************************************************************
    ! The error in the corrate
    !**********************************************************************
    corrRateErr = (abs(corrRateErrMinus1) + abs(corrRateErrPlus1)) / 2.0  
    !write(*,*) "L466 corrRateErr=", corrRateErr

    !************************************************************
    ! Correct the background for coincidence-loss correction
    !************************************************************
    oldCorrrate = source%corrBackgroundCountRate

    !************************************************************
    ! 2013-10-23 the input error parameter is not giving the value of the
    ! background error. Instead it is the source error (fixing)
    ! *****************************************************************
    error  = source%corrBackGroundCountRateErr
    !write(*,*) "L478 CL-aperture background error=", error
    source%corrBackgroundCountRate = CorrectForCoincidenceLoss(dble(source%corrBackGroundCountRate), &
                     error, frameTime, corrRateErrMinus2, corrRateErrPlus2)     
    corrBkgRateErr = (abs(corrRateErrMinus2) + abs(corrRateErrPlus2)) / 2.0  
    !write(*,*) "L482 corrBackgroundCountRate=", source%corrBackgroundCountRate, " corrRateErrMinus2=", &
    !     corrRateErrMinus2, &
    !     " errPlus=", corrRateErrPlus2
    
    new2oldBkgCorrRate=1.0_double
    if (oldCorrrate  .gt. 0.0) then
       new2oldBkgCorrRate=source%corrBackgroundCountRate/oldCorrrate
    endif
    !write(*,*) "L490 new2oldBkgCorrRate=", new2oldBkgCorrRate
    !***************************************************************************
    ! 2013-11-23 calculate the factor of the background error increase due to the
    ! coincidence-loss correction
    !****************************************************************************
    !new2oldBkgCorrRateErr=1.0_double
    !if (error  .gt. 0.0) then
    !   new2oldBkgCorrRateErr=corrBkgRateErr/error
    !endif
    !write(*,*) "L446 new2oldBkgCorrRateErr=", new2oldBkgCorrRateErr
    if (oldcorrrate .gt. 0.0) then
       source%declination = source%corrBackgroundCountRate / oldCorrRate
    else
       source%declination = 0.0
    end if

    !write(*,*) "L506 source corrBackgroundCountRate =", source%corrBackgroundCountRate
    !*****************************************
    ! Subtract the corrected background
    !***************************************** 
    oldCorrRate=source%corrCountRate
    source%corrCountRate = corrTotalRate - source%corrBackgroundCountRate   
    !write(*,*) "L512 source corrCountRate=", source%corrCountRate
    aux=source%corrCountRate/s2nratio
    !write(*,*) "L514 aux=", aux
    new2oldRawRate=1.0_double
    if (oldCorrRate .gt. 0.0) then
       new2oldRawRate=source%corrCountRate/oldCorrRate
    endif
    !write(*,*) "L517 new2oldRawRate=", new2oldRawRate

    !*****************************************************************************
    ! 2013-11-23 we have to re-scale the background rate and its error 
    ! to the measuring aperture
    !*****************************************************************************
    ! 
    source%corrBackgroundCountRate=source%rawBackGroundCountRate*new2oldBkgCorrRate
    !**********************************************************
    ! The error in the background count-rate
    !**********************************************************
    !source%corrBackgroundCountRateErr = (abs(corrRateErrMinus2) + abs(corrRateErrPlus2)) / 2.0    
    !source%corrBackgroundCountRateErr=source%rawBackGroundCountRateErr*new2oldBkgCorrRateErr
    ! 2013-11-26 correcting the mistake in scaling the background error
    source%corrBackgroundCountRateErr=source%rawBackGroundCountRateErr*new2oldBkgCorrRate

    !******************************************************
    ! The error in the background-subtracted count rate
    !****************************************************** 
    ! 2013-10-01 vny: corrRateErr already includes the background.
    ! Why should we add the background error again? This might be a mistake.


    aux=source%corrCountRateErr


    !write(*,*) "L543 corrCountRateErr=", source%corrCountRateErr
   ! 2013-11-26 correcting a mistake in the scaling of the source error:
    !***************************************************************
    source%corrCountRateErr=aux*new2oldRawRate
    !write(*,*) "L547 test corrCountRateErr=", source%corrCountRateErr, " new2oldRawRate=", &
    !     new2oldRawRate
    !***************************************************************

    !**********************************************************************
    !write(*,*) "L552 source corrCountRateErr=", aux, &
    !     " new2oldTotRate=", new2oldTotRate
    source%corrCountRateErr=aux*new2oldTotRate    
    !write(*,*) "L555 corrCountRateErr=", source%corrCountRateErr, " new2oldTotRate=", &
    !     new2oldTotRate

    !ttttttttttttttttttt
    source%corrCountRateErr=source%corrCountRate/s2nRatio    
    !write(*,*) "L558 corrCountRateErr=", source%corrCountRateErr, " s2nratio=", &
    !     s2nratio
    !tttttttttttttttttttt

    !write(*,*) "----"

    !*********************************************
    ! 2013-11-23 vny: here we have to introduce the rigorous CL-correctied 
    ! rate error calculation by using the Simon&Paul 2008 formula
    !********************************************************************
    ! NBNBNBNBNB
    !*******************************************************************

end subroutine PointSourceCoincLossCorr
!************************************************************
! Function PSFCountsRatio
! Computes the ratio of the total counts between
! the radii radius1 and radius2 
!************************************************************
real(double) Function PSFCountsRatio(radius1, radius2, x0, y0, countFrameRatio)
    real(single), intent(in) :: radius1    ! The aperture radius that the input counts refer to
    real(single), intent(in) :: radius2    ! The radius that the counts have to be scaled to 
    real(single), intent(in) :: x0         ! The x coordinate of the source 
    real(single), intent(in) :: y0         ! The y coordinate of the source 

    real(single), intent(in) :: countFrameRatio

    PSFCountsRatio = PSFFractionOfTotalCounts(x0, y0, radius1, countFrameRatio) / &
                     PSFFractionOfTotalCounts(x0, y0, radius2, countFrameRatio)


end function PSFCountsRatio  

!*********************************************************
! Function CorrectForCoincidenceLoss
! Corrects the raw count rate for coincidence losses  
!*********************************************************
real(double) function CorrectForCoincidenceLoss(rawCountRate, & 
     countRateError, frameTime,  &
     countRateErrorMinus, countRateErrorPlus)
  real(double), intent(in)     :: rawCountRate
  real(single), intent(in)     :: countRateError
  real(double), intent(out)    :: countRateErrorMinus
  real(double), intent(out)    :: countRateErrorPlus
  
  real(double), intent(in) :: frameTime
  real(single) :: rawTempRate
  type(ValueAsymmetricErrorT) :: rawRate
  type(ValueAsymmetricErrorT) :: corrRate


  
  if (rawCountRate .le. 0.0_double .or. isNullValue(rawCountrate) ) then
     !write(*,*) "L471 exiting CorrectForCoincidenceLoss countRateError=", countRateError
     CorrectForCoincidenceLoss = rawCountRate
     countRateErrorMinus = countRateError
     countRateErrorPlus = countRateError
     return
  end if
  !write(*,*) "L477 frameTime=", frameTime
  if(abs(frameTime) < 1e-6) then
     !write(*,*) "L479 exiting CorrectForCoincidenceLoss countRateError=", countRateError
     CorrectForCoincidenceLoss = rawCountRate
     countRateErrorMinus = countRateError
     countRateErrorPlus  = countRateError
     return
  end if
  
  !**************************************************************
  ! 2013-11-15 vny: the error bar for the coincidence-loss corrected errors
  ! is growing tremendously when the rawCountRate approaches the value 1/frameTime.
  ! This make no much sense, so we can limit the error bar by the value of
  ! the rawCountRate. So, instead of the constant 0.99999 we shall use
  ! the constant 0.99 which should improve the situation
  !***************************************************************


  !*************************************************************
  ! NB frame time is in seconds
  !*************************************************************
  ! 2013-11-15 changing the constant from 0.99999 to 0.99 to avoid
  ! getting nonsense values for error bars 
  !*************************************************************
  !rawTempRate = min(0.99999_double / frameTime, dble(rawCountRate)) 
  rawTempRate = min(0.99_double / frameTime, dble(rawCountRate)) 
  !write(*,*) "L491 rawTempRate=", rawTemprate
  rawRate = ValueAsymmetricErrorT(rawTempRate, dble(countRateError), dble(countRateError))
  !write(*,*) "L493 rawRate=", rawRate, " frameTime=", frameTime
  
  !******************************************************************************************
  !
  ! NB the final boolean parameter switches on the empirical correction
  !
  !*****************************************************************************************
  corrRate = CAL_omPhotoNatural(rawRate, frameTime, 0.0_double, 0.0_double, .FALSE., .TRUE.)
  !write(*,*) "L501 corrRate%value=", corrRate%value, " corrRate%errorMinus=", corrRate%errorMinus
  !****************************************************************
  ! If the count rate is greater than the frame rate then the above
  ! routine cannot do the correction- instead it sets the corrected 
  ! count rate to the largest number that can be stored on the computer. 
  !*****************************************************************
  
  if (corrRate%value .lt. 1e10) then
     !write(*,*) "L509 corrRate%errorMinus=", corrRate%errorMinus  
     CorrectForCoincidenceLoss = real(CorrRate%value)        
     countRateErrorMinus = corrRate%errorMinus     
     countRateErrorPlus  = corrRate%errorPlus        
  else
     !write(*,*) "L514 rawCountRate=", rawCountRate
     CorrectForCoincidenceLoss = rawCountRate       
  end if
  
end function CorrectForCoincidenceLoss

!******************************************
! PSFFractionOfTotalCounts
! Returns the fraction of the total counts
! contained within the given radius
!******************************************
real(double) function & 
    PSFFractionOfTotalCounts(x0, y0, radius, countFrameRatio)
    real(single), intent(in) :: x0
    real(single), intent(in) :: y0
    real(single), intent(in) :: radius
    real(single), intent(in) :: countFrameRatio
    type(OmPsfT) :: psf
 
    psf = CAL_omGetPSF(dble(x0), dble(y0), dble(countFrameRatio))

    PSFFractionOfTotalCounts = CAL_omPsfEncircledEnergy(psf, dble(radius))


end function PSFFractionOfTotalCounts 

end module omcoinccorr_mod


























