! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
module omfunctions_mod
use dal
use omallocation_mod

!***********************************************************************
!
!
!***********************************************************************
type FactorialStruct
    integer :: maxSize = 5000
    logical, pointer :: computed(:) => null()
    real(single), pointer :: computedValue(:) => null()


end type FactorialStruct
type(FactorialStruct), save :: factorialData 


contains

!***********************************************************************
!
!
!     ALGORITHM AS245  APPL. STATIST. (1989) VOL. 38, NO. 2
!
!     Calculation of the logarithm of the gamma function
!
! ELF90-compatible version by Alan Miller
! Latest revision - 29 November 1997
!
! N.B. Argument IFAULT has been removed
!
!
!***********************************************************************
FUNCTION logGamma(xvalue) RESULT(fn_val)


IMPLICIT NONE

REAL (double), INTENT(IN) :: xvalue
REAL (double)             :: fn_val

! Local variables
REAL (double) :: x, x1, x2, y

!     Coefficients of rational functions

REAL (double), PARAMETER :: r1(9) = (/ -2.66685511495_double, -24.4387534237_double,  &
                                   -21.9698958928_double,  11.1667541262_double,  &
                                    3.13060547623_double,  0.607771387771_double, &
                                    11.9400905721_double,  31.4690115749_double,  &
                                    15.2346874070_double /)
REAL (double), PARAMETER :: r2(9) = (/ -78.3359299449_double, -142.046296688_double,  &
                                    137.519416416_double,  78.6994924154_double,  &
                                    4.16438922228_double,  47.0668766060_double,  &
                                    313.399215894_double,  263.505074721_double,  &
                                    43.3400022514_double /)
REAL (double), PARAMETER :: r3(9) = (/ -2.12159572323E5_double,  2.30661510616E5_double,  &
                                    2.74647644705E4_double, -4.02621119975E4_double,  &
                                   -2.29660729780E3_double, -1.16328495004E5_double,  &
                                   -1.46025937511E5_double, -2.42357409629E4_double,  &
                                   -5.70691009324E2_double /)
REAL (double), PARAMETER :: r4(5) = (/ 0.279195317918525_double, 0.4917317610505968_double, &
                                   0.0692910599291889_double, 3.350343815022304_double, &
                                   6.012459259764103_double /)

!     Fixed constants

REAL (double), PARAMETER :: alr2pi = 0.918938533204673_double, four = 4.0_double,  &
                        half = 0.5_double, one = 1._double, onep5 = 1.5_double,   &
                        twelve = 12._double, zero = 0._double

!     Machine-dependant constants.
!     A table of values is given at the top of page 399 of the paper.
!     These values are for the IEEE double-precision format for which
!     B = 2, t = 53 and U = 1023 in the notation of the paper.

REAL (double), PARAMETER :: xlge = 5.10E6_double, xlgst = HUGE(1.0_double)


!write(*, *) "LogGamma- xvalue=", xvalue

x = xvalue
fn_val = zero

!     Test for valid function argument

IF (x >= xlgst) THEN
  WRITE(*, *) 'AS 245: Argument x too large'
  RETURN
END IF
IF (x <= zero) THEN
 
  RETURN
END IF

!     Calculation for 0 < X < 0.5 and 0.5 <= X < 1.5 combined

IF (x < onep5) THEN
  IF (x < half) THEN
    fn_val = -LOG(x)
    y = x + one
    
!     Test whether X < machine epsilon
    
    IF (y == one) RETURN
  ELSE
    fn_val = zero
    y = x
    x = (x - half) - half
  END IF
  fn_val = fn_val + x * ((((r1(5)*y + r1(4))*y + r1(3))*y + r1(2))*y + r1(1)) / &
                    ((((y + r1(9))*y + r1(8))*y+ r1(7))*y + r1(6))
  RETURN
END IF

!     Calculation for 1.5 <= X < 4.0

IF (x < four) THEN
  y = (x - one) - one
  fn_val = y * ((((r2(5)*x + r2(4))*x + r2(3))*x + r2(2))*x + r2(1)) /  &
               ((((x + r2(9))*x + r2(8))*x + r2(7))*x+ r2(6))
  RETURN
END IF

!     Calculation for 4.0 <= X < 12.0

IF (x < twelve) THEN
  fn_val = ((((r3(5)*x + r3(4))*x + r3(3))*x + r3(2))*x + r3(1)) /  &
           ((((x + r3(9))*x + r3(8))*x + r3(7))*x + r3(6))
  RETURN
END IF

!     Calculation for X >= 12.0

y = LOG(x)
!write(*, *) "LogGamma Line 127 x=", x

fn_val = x * (y - one) - half * y + alr2pi
IF (x > xlge) RETURN
!write(*, *) "LogGamma Line 131 x=", x
x1 = one / x
x2 = x1 * x1
fn_val = fn_val + x1 * ((r4(3)*x2 + r4(2))*x2 + r4(1)) /  &
         ((x2 + r4(5))*x2 + r4(4))


END FUNCTION logGamma



!***********************************************
!
!***********************************************
real(double) Function gammap(a, x)
    real(double), intent(in) :: a, x
    real(double) :: gln

    if ( x .lt. (a + 1.0_double)) then
        Gammap = gser(a, x, gln)

    else
        Gammap = 1.0_double - gcf(a, x, gln)
    end if

end function Gammap

!***********************************************
!
!***********************************************
real(double) Function gammaq(a, x)
    real(double), intent(in) :: a, x
    real(double) :: gln

    if ( x .lt. a + 1.0_double) then
        Gammaq = 1.0_double - gser(a, x, gln)

    else
        Gammaq = gcf(a, x, gln)
    end if
end function Gammaq
!***********************************************
!
!***********************************************
real(double) Function gser(a, x, gln)
    real(double), intent(in) :: a, x
    real(double), intent(out) :: gln

    integer :: n
    integer, parameter :: ITMAX = 2000

    real(double) :: sum, del, ap, ap1
    real(double), parameter :: EPS = 3.0e-7

    if (x .le. 0.0_double) then
        Gser = 0.0_double
        return
    end if

    gln = logGamma(a)

    ap1 = a
    del = 1.0_double / a
    sum = del
    do n = 1, ITMAX
        ap = ap1 + dble(n)
        del = del * x / ap
        sum = sum + del
        if (abs(del) .lt. abs(sum) * EPS) then
            gser = sum * exp(-x + a * log(x) - gln)
            return
        end if
    end do
!write(*, *) "a too large, ITMAX too small in gser (a=", a, " x=", x, ")"
end function Gser
!***********************************************
!
!***********************************************
real(double) Function gcf(a, x, gln)
    real(double), intent(in) :: a, x
    real(double), intent(out) :: gln
    integer :: i
    real(double) :: an, b, c, d, del, h
    integer, parameter :: ITMAX = 2000

   
    real(double), parameter :: EPS = 3.0e-7
    real(double), parameter :: FPMIN = 1.0e-30


    gln = logGamma(a)
    b = x + 1.0_double - a
    c = 1.0_double / FPMIN
    d = 1.0_double / b
    h = d
    do i = 1, ITMAX
        an = -dble(i) * (dble(i) - a)
        b = b + 2.0_double
        d = an * d + b
        if (abs(d) .lt. FPMIN) d = FPMIN
        c = b + an / c
        if (abs(c) .lt. FPMIN) c = FPMIN
        d = 1.0_double / d
        del = d * c
        h = h * del
        if (abs(del - 1.0_double) .lt. EPS) exit
    end do
    if (i .gt. ITMAX) write(*, *) "a too large, ITMAX too small in gcf"
    gcf = exp(-x + a * log(x) - gln) * h

end function gcf





!******************************************************************
!
!
!******************************************************************

subroutine InitialiseFactorial
    if (associated(factorialData%computed)) return
    factorialData%maxSize = 10000
    allocate(factorialData%computed(0 : factorialData%maxSize))
    factorialData%computed = .false.

    factorialData%computedValue => allocateSingle(0, factorialData%maxSize, "computedValue")


end subroutine InitialiseFactorial

!******************************************************************
!
!
!******************************************************************
subroutine DestructFactorial
    deallocate(factorialData%computed)
    deallocate(factorialData%computedValue)

    nullify(factorialData%computed)
    nullify(factorialData%computedValue)


end subroutine DestructFactorial

!*******************************************
!
!******************************************
real(single) Function logFactorial(i)
    integer, intent(in) :: i
    integer(int32) :: n, j

    if (i .le. factorialData%maxSize) then
        if (factorialData%computed(i)) then
            logFactorial = factorialData%computedValue(i)

        else
            if (i .lt. 10) then
                if (i .eq. 0) then
                    logFactorial = 0.0
                   
                else
                   n = 1 
                   do j = 1, i
                       n = n * j
                   end do
                   logFactorial = log(real(n))
                end if 
            else    
                logFactorial = logGamma(dble(i + 1))

 
            end if
            factorialData%computed(i) = .true.
            factorialData%computedValue(i) = logFactorial
            
        end if
    else
        logFactorial = logGamma(dble(i + 1))

    end if
 
end Function logFactorial


end module omfunctions_mod
