! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
module omMinim_mod
  use dal
  
use errorHandling
 


PRIVATE
PUBLIC :: omminim

CONTAINS

!************************************************************************
!     A PROGRAM FOR FUNCTION MINIMIZATION USING THE SIMPLEX METHOD.

!     FOR DETAILS, SEE NELDER & MEAD, THE COMPUTER JOURNAL, JANUARY 1965

!     PROGRAMMED BY D.E.SHAW,
!     CSIRO, DIVISION OF MATHEMATICS & STATISTICS
!     P.O. BOX 218, LINDFIELD, N.S.W. 2070

!     WITH AMENDMENTS BY R.W.M.WEDDERBURN
!     ROTHAMSTED EXPERIMENTAL STATION
!     HARPENDEN, HERTFORDSHIRE, ENGLAND

!     Further amended by Alan Miller
!     CSIRO Division of Mathematical & Information Sciences
!     Private Bag 10, CLAYTON, VIC. 3169

!     Fortran 90 conversion by Alan Miller, June 1995
!     Alan.Miller @ vic.cmis.csiro.au
!     Latest revision - 5 December 1999

!     ARGUMENTS:-
!     P()     = INPUT, STARTING VALUES OF PARAMETERS
!               OUTPUT, FINAL VALUES OF PARAMETERS
!     STEP()  = INPUT, INITIAL STEP SIZES (if zero fixed variable)
!     NOP     = INPUT, NO. OF PARAMETERS, INCL. ANY TO BE HELD FIXED
!     FUNC    = OUTPUT, THE FUNCTION VALUE CORRESPONDING TO THE FINAL
!                 PARAMETER VALUES.
!     maxfn     = INPUT, THE MAXIMUM NO. OF FUNCTION EVALUATIONS ALLOWED.
!               Say, 20 times the number of parameters, NOP.
!     STOPCR  = INPUT, STOPPING CRITERION.
!               The criterion is applied to the standard deviation of
!               the values of FUNC at the points of the simplex.
!     NLOOP   = INPUT, THE STOPPING RULE IS APPLIED AFTER EVERY NLOOP
!               FUNCTION EVALUATIONS.   Normally NLOOP should be slightly
!               greater than NOP, say NLOOP = 2*NOP.
!     IQUAD   = INPUT, = 1 IF FITTING OF A QUADRATIC SURFACE IS REQUIRED
!                      = 0 IF NOT
!               N.B. The fitting of a quadratic surface is strongly
!               recommended, provided that the fitted function is
!               continuous in the vicinity of the minimum.   It is often
!               a good indicator of whether a premature termination of
!               the search has occurred.
!     SIMP    = INPUT, CRITERION FOR EXPANDING THE SIMPLEX TO OVERCOME
!               ROUNDING ERRORS BEFORE FITTING THE QUADRATIC SURFACE.
!               The simplex is expanded so that the function values at
!               the points of the simplex exceed those at the supposed
!               minimum by at least an amount SIMP.
!     VAR()   = OUTPUT, CONTAINS THE DIAGONAL ELEMENTS OF THE INVERSE OF
!               THE INFORMATION MATRIX.
!     FUNCTN  = INPUT, NAME OF THE USER'S SUBROUTINE - ARGUMENTS (P,FUNC)
!               WHICH RETURNS THE FUNCTION VALUE FOR A GIVEN SET OF
!               PARAMETER VALUES IN ARRAY P.
!****     FUNCTN MUST BE DECLARED EXTERNAL IN THE CALLING PROGRAM.
!     IFAULT  = OUTPUT, = 0 FOR SUCCESSFUL TERMINATION
!                 = 1 IF MAXIMUM NO. OF FUNCTION EVALUATIONS EXCEEDED
!                 = 2 IF INFORMATION MATRIX IS NOT +VE SEMI-DEFINITE
!                 = 3 IF NOP < 1
!                 = 4 IF NLOOP < 1

!     N.B. P, STEP AND VAR (IF IQUAD = 1) MUST HAVE DIMENSION AT LEAST NOP
!          IN THE CALLING PROGRAM.

!*****************************************************************************

SUBROUTINE omminim(p, step1, nop, func, maxfn, stopcr, nloop, iquad,  &
                 simp, var, functn, ifault)


INTEGER, INTENT(IN)        :: nop, maxfn, nloop, iquad
INTEGER, INTENT(OUT)       :: ifault
REAL(double), INTENT(IN)      :: stopcr, simp
REAL(double), INTENT(IN OUT)  :: p(:)
REAL(double), INTENT(IN)  :: step1(:)

REAL(double), INTENT(OUT)     :: var(:), func

INTERFACE

!*********************************************************
! The function to be minimised
!*********************************************************
Function functn(p)
use dal
    IMPLICIT NONE
       
    REAL (double), INTENT(inout)  :: p(:)
    real(double) :: functn
        
  END Function functn
END INTERFACE

!x     Local variables

REAL (double)   :: g(nop+1,nop), h(nop+1), pbar(nop), pstar(nop), pstst(nop), step(nop), &
               aval(nop), pmin(nop), temp(nop), bmat(nop*(nop+1)/2),  &
               vc(nop*(nop+1)/2), ymin, rmax, hstst, a0, a03, hmin, test, hmean, &
               hstd, hstar, hmax, savemn

REAL(double), PARAMETER :: zero = 0._double, half = 0.5_double, one = 1._double, two = 2._double
real(double), parameter :: tolerance = 10.0_double * epsilon(1.0_double)
INTEGER     :: i, i1, i2, iflag, imax, imin, irank, irow, j, j1, &
               k, l, loop, nap, neval, nmore, np1, nullty


!     A = REFLECTION COEFFICIENT, B = CONTRACTION COEFFICIENT, AND
!     C = EXPANSION COEFFICIENT.

REAL (double), PARAMETER :: a = 1._double, b = 0.5_double, c = 2._double


logical :: toFit(nop)
!    write(errstr,'(a)') "Minimization subroutine minim"
!    call message(AppMsg, Verbose)




!     CHECK INPUT ARGUMENTS
    step(1 : nop) = step1(1 : nop)
    ifault = 0
    IF (nop <= 0) ifault = 3
    IF (nloop <= 0) ifault = 4
    IF (ifault /= 0) RETURN

!     SET NAP = NO. OF PARAMETERS TO BE VARIED, I.E. WITH STEP /= 0
    where(step == 0.0)
        toFit = .false.
    elsewhere
        toFit = .true.
    end where

    nap = COUNT(step /= zero)
    neval = 0
    loop = 0
    iflag = 0

!     IF NAP = 0 EVALUATE FUNCTION AT THE STARTING POINT AND RETURN

    IF (nap <= 0) THEN
        func = functn(p)
        RETURN
    END IF

!     SET UP THE INITIAL SIMPLEX

20 g(1, :) = p
    irow = 2
    DO i = 1, nop
        IF (toFit(i)) THEN
            g(irow, :) = p
            g(irow, i) = p(i) + step(i)
            irow = irow + 1
        END IF
    END DO

    np1 = nap + 1
    DO i = 1, np1
        p = g(i, :)
        h(i) =  functn(p)
        neval = neval + 1
  
    END DO

!     START OF MAIN CYCLE.

!     FIND MAX. & MIN. VALUES FOR CURRENT SIMPLEX (HMAX & HMIN).

Main_loop: DO
   

    loop = loop + 1
!  write(errstr,'(a, i6)') "Main loop loop number=", loop
!    call message(AppMsg, Verbose)

    imax = 1
    imin = 1
    hmax = h(1)
    hmin = h(1)
    DO i = 2, np1
        IF (h(i) > hmax) THEN
            imax = i
            hmax = h(i)
        ELSE
            IF (h(i) < hmin) THEN
                imin = i
                hmin = h(i)
            END IF
        END IF
    END DO

!     FIND THE CENTROID OF THE VERTICES OTHER THAN P(IMAX)

    pbar = zero
    DO i = 1, np1
        IF (i /= imax) THEN
            pbar = pbar + g(i, :)
        END IF
    END DO
    pbar = pbar / nap

!     REFLECT MAXIMUM THROUGH PBAR TO PSTAR,
!     HSTAR = FUNCTION VALUE AT PSTAR.

    pstar = a * (pbar - g(imax,:)) + pbar
    hstar = functn(pstar)
    neval = neval + 1
 

!     IF HSTAR < HMIN, REFLECT PBAR THROUGH PSTAR,
!     HSTST = FUNCTION VALUE AT PSTST.

    IF (hstar < hmin) THEN
        pstst = c * (pstar - pbar) + pbar
        hstst =  functn(pstst)
        neval = neval + 1
    

!     IF HSTST < HMIN REPLACE CURRENT MAXIMUM POINT BY PSTST AND
!     HMAX BY HSTST, THEN TEST FOR CONVERGENCE.

        IF (hstst >= hmin) THEN   ! REPLACE MAXIMUM POINT BY PSTAR & H(IMAX) BY HSTAR.
            g(imax, :) = pstar
            h(imax) = hstar
        ELSE
            g(imax, :) = pstst
            h(imax) = hstst
        END IF
        GO TO 250
    END IF

!     HSTAR IS NOT < HMIN.
!     TEST WHETHER IT IS < FUNCTION VALUE AT SOME POINT OTHER THAN
!     P(IMAX).   IF IT IS REPLACE P(IMAX) BY PSTAR & HMAX BY HSTAR.

    DO i = 1, np1
        IF (i /= imax) THEN
            IF (hstar < h(i)) THEN  ! REPLACE MAXIMUM POINT BY PSTAR & H(IMAX) BY HSTAR.
                g(imax, :) = pstar
                h(imax) = hstar
                GO TO 250
            END IF
        END IF
    END DO

!     HSTAR > ALL FUNCTION VALUES EXCEPT POSSIBLY HMAX.
!     IF HSTAR <= HMAX, REPLACE P(IMAX) BY PSTAR & HMAX BY HSTAR.

    IF (hstar <= hmax) THEN
        g(imax, :) = pstar
        hmax = hstar
        h(imax) = hstar
    END IF

!     CONTRACTED STEP TO THE POINT PSTST,
!     HSTST = FUNCTION VALUE AT PSTST.

    pstst = b * g(imax, :) + b * pbar
    hstst = functn(pstst)
    neval = neval + 1
  

!     IF HSTST < HMAX REPLACE P(IMAX) BY PSTST & HMAX BY HSTST.

    IF (hstst <= hmax) THEN
        g(imax, :) = pstst
        h(imax) = hstst
        GO TO 250
    END IF

!     HSTST > HMAX.
!     SHRINK THE SIMPLEX BY REPLACING EACH POINT, OTHER THAN THE CURRENT
!     MINIMUM, BY A POINT MID-WAY BETWEEN ITS CURRENT POSITION AND THE
!     MINIMUM.

    DO i = 1, np1
        IF (i /= imin) THEN
            DO j = 1, nop
                IF (toFit(j)) g(i, j) = (g(i, j) + g(imin, j)) * half
                p(j) = g(i, j)
            END DO
            h(i) =  functn(p)
            neval = neval + 1
    
        END IF
    END DO

!     IF LOOP = NLOOP TEST FOR CONVERGENCE, OTHERWISE REPEAT MAIN CYCLE.

 
  250 IF (loop < nloop) CYCLE Main_loop

!     CALCULATE MEAN & STANDARD DEVIATION OF FUNCTION VALUES FOR THE
!     CURRENT SIMPLEX.

    hmean = SUM( h(1 : np1) ) / np1
    hstd = SUM( (h(1 : np1) - hmean) ** 2 )
    hstd = SQRT(hstd / np1)

!     IF THE RMS > STOPCR, SET IFLAG & LOOP TO ZERO AND GO TO THE
!     START OF THE MAIN CYCLE AGAIN.



    IF (hstd > stopcr .AND. neval <= maxfn) THEN
        iflag = 0
        loop = 0
        CYCLE Main_loop
      END IF

!     FIND THE CENTROID OF THE CURRENT SIMPLEX AND THE FUNCTION VALUE THERE.

    DO i = 1, nop
        IF (toFit(i)) THEN
            p(i) = SUM( g(1 : np1, i) ) / np1
        END IF
    END DO
    func = functn(p)
    neval = neval + 1
 

!     TEST WHETHER THE NO. OF FUNCTION VALUES ALLOWED, maxfn, HAS BEEN
!     OVERRUN; IF SO, EXIT WITH IFAULT = 1.

    IF (neval > maxfn) THEN
        ifault = 1
    
        RETURN
    END IF

!     CONVERGENCE CRITERION SATISFIED.
!     IF IFLAG = 0, SET IFLAG & SAVE HMEAN.
!     IF IFLAG = 1 & CHANGE IN HMEAN <= STOPCR THEN SEARCH IS COMPLETE.


    IF (iflag == 0 .OR. ABS(savemn - hmean) >= stopcr) THEN
        iflag = 1
        savemn = hmean
        loop = 0
    ELSE
        EXIT Main_loop
    END IF

    END DO Main_loop

    IF (iquad <= 0) RETURN

!------------------------------------------------------------------

!     QUADRATIC SURFACE FITTING



!     EXPAND THE FINAL SIMPLEX, IF NECESSARY, TO OVERCOME ROUNDING
!     ERRORS.

    hmin = func
    nmore = 0
    DO i = 1, np1
        DO
            test = ABS(h(i) - func)
            IF (test < simp) THEN
            DO j = 1, nop
                IF (toFit(j)) g(i, j) = (g(i, j) - p(j)) + g(i, j)
                pstst(j) = g(i, j)
            END DO
            h(i) = functn(pstst)
            nmore = nmore + 1
            neval = neval + 1
            IF (h(i) >= hmin) CYCLE
            hmin = h(i)
      
        ELSE
            EXIT
        END IF
    END DO
END DO

!     FUNCTION VALUES ARE CALCULATED AT AN ADDITIONAL NAP POINTS.

    DO i = 1, nap
        i1 = i + 1
        pstar = (g(1, :) + g(i1, :)) * half
        aval(i) =  functn(pstar)
        nmore = nmore + 1
        neval = neval + 1
    END DO

!     THE MATRIX OF ESTIMATED SECOND DERIVATIVES IS CALCULATED AND ITS
!     LOWER TRIANGLE STORED IN BMAT.

    a0 = h(1)
    a03 = 3.0_double * a0
    DO i = 1, nap
        i1 = i - 1
        i2 = i + 1
        DO j = 1, i1
            j1 = j + 1
            pstst = (g(i2, :) + g(j1, :)) * half
            hstst = functn(pstst)
            nmore = nmore + 1
            neval = neval + 1
            l = i * (i-1) / 2 + j
            bmat(l) = two * (hstst + a0 - aval(i) - aval(j))
        END DO
    END DO

    l = 0
    DO i = 1, nap
        i1 = i + 1
        l = l + i
        bmat(l) = two * (h(i1) + a0 - two * aval(i))
    END DO

!     THE VECTOR OF ESTIMATED FIRST DERIVATIVES IS CALCULATED AND
!     STORED IN AVAL.

    DO i = 1, nap
        i1 = i + 1
        aval(i) = two * aval(i) - (h(i1) + a03) * half
    END DO

!     THE MATRIX Q OF NELDER & MEAD IS CALCULATED AND STORED IN G.

    pmin = g(1, :)
    DO i = 1, nap
        i1 = i + 1
        g(i1, :) = g(i1, :) - g(1, :)
    END DO

    DO i = 1, nap
        i1 = i + 1
        g(i, :) = g(i1, :)
    END DO

!     INVERT BMAT

    CALL syminv(bmat, nap, bmat, temp, nullty, ifault, rmax)
    IF (ifault == 0) THEN
        irank = nap - nullty
    ELSE                                 ! BMAT not +ve definite
                                     ! Resume search for the minimum
  
        ifault = 2
        IF (neval > maxfn) RETURN
        do i = 1, nop
            if (step(i) .ne. 0.0) then        
                step(i) = half * step(i)
                if (abs(step(i)) <= tolerance) return 
            end if
        end do
        
        GO TO 20
    END IF

!     BMAT*A/2 IS CALCULATED AND STORED IN H.

    DO i = 1, nap
        h(i) = zero
        DO j = 1, nap
            IF (j <= i) THEN
                l = i * (i - 1) / 2 + j
             ELSE
                 l = j * (j - 1) / 2 + i
             END IF
                 h(i) = h(i) + bmat(l) * aval(j)
        END DO
    END DO

!     FIND THE POSITION, PMIN, & VALUE, YMIN, OF THE MINIMUM OF THE
!     QUADRATIC.

    ymin = DOT_PRODUCT( h(1 : nap), aval(1 : nap) )
    ymin = a0 - ymin
    DO i = 1, nop
        pstst(i) = DOT_PRODUCT( h(1 : nap), g(1 : nap, i) )
    END DO
    pmin = pmin - pstst


!     Q*BMAT*Q'/2 IS CALCULATED & ITS LOWER TRIANGLE STORED IN VC

    DO i = 1, nop
        DO j = 1, nap
            h(j) = zero
            DO k = 1, nap
                IF (k <= j) THEN
                    l = j * (j - 1) / 2 + k
                ELSE
                    l = k * (k - 1) / 2 + j
                END IF
                h(j) = h(j) + bmat(l) * g(k,i) * half
            END DO
        END DO

    DO j = i, nop
        l = j * (j - 1) / 2 + i
        vc(l) = DOT_PRODUCT( h(1 : nap), g(1 : nap, j) )
    END DO
    END DO

!     THE DIAGONAL ELEMENTS OF VC ARE COPIED INTO VAR.

    j = 0
    DO i = 1, nop
        j = j + i
            var(i) = vc(j)
    END DO

    RETURN

END SUBROUTINE omminim




SUBROUTINE syminv(a, n, c, w, nullty, ifault, rmax)

!     ALGORITHM AS7, APPLIED STATISTICS, VOL.17, 1968.

!     ARGUMENTS:-
!     A()    = INPUT, THE SYMMETRIC MATRIX TO BE INVERTED, STORED IN
!                LOWER TRIANGULAR FORM
!     N      = INPUT, ORDER OF THE MATRIX
!     C()    = OUTPUT, THE INVERSE OF A (A GENERALIZED INVERSE IF C IS
!                SINGULAR), ALSO STORED IN LOWER TRIANGULAR FORM.
!                C AND A MAY OCCUPY THE SAME LOCATIONS.
!     W()    = WORKSPACE, DIMENSION AT LEAST N.
!     NULLTY = OUTPUT, THE RANK DEFICIENCY OF A.
!     IFAULT = OUTPUT, ERROR INDICATOR
!                 = 1 IF N < 1
!                 = 2 IF A IS NOT +VE SEMI-DEFINITE
!                 = 0 OTHERWISE
!     RMAX   = OUTPUT, APPROXIMATE BOUND ON THE ACCURACY OF THE DIAGONAL
!                ELEMENTS OF C.  E.G. IF RMAX = 1.E-04 THEN THE DIAGONAL
!                ELEMENTS OF C WILL BE ACCURATE TO ABOUT 4 DEC. DIGITS.

!     LATEST REVISION - 1 April 1985

!***************************************************************************

REAL (double), INTENT(IN OUT) :: a(:), c(:), w(:)
INTEGER, INTENT(IN)       :: n
INTEGER, INTENT(OUT)      :: nullty, ifault
REAL (double), INTENT(OUT)    :: rmax

REAL (double), PARAMETER :: zero = 0._double, one = 1._double
INTEGER              :: i, icol, irow, j, jcol, k, l, mdiag, ndiag, nn, nrow
REAL (double)            :: x

nrow = n
ifault = 1
IF (nrow > 0) THEN
  ifault = 0

!     CHOLESKY FACTORIZATION OF A, RESULT IN C

  CALL chola(a, nrow, c, nullty, ifault, rmax, w)
  IF (ifault == 0) THEN

!     INVERT C & FORM THE PRODUCT (CINV)'*CINV, WHERE CINV IS THE INVERSE
!     OF C, ROW BY ROW STARTING WITH THE LAST ROW.
!     IROW = THE ROW NUMBER, NDIAG = LOCATION OF LAST ELEMENT IN THE ROW.

    nn = nrow * (nrow+1) / 2
    irow = nrow
    ndiag = nn
    10 IF (c(ndiag) /= zero) THEN
      l = ndiag
      DO i = irow, nrow
        w(i) = c(l)
        l = l + i
      END DO
      icol = nrow
      jcol = nn
      mdiag = nn

      30 l = jcol
      x = zero
      IF (icol == irow) x = one / w(irow)
      k = nrow
      40 IF (k /= irow) THEN
        x = x - w(k) * c(l)
        k = k - 1
        l = l - 1
        IF (l > mdiag) l = l - k + 1
        GO TO 40
      END IF

      c(l) = x / w(irow)
      IF (icol == irow) GO TO 60
      mdiag = mdiag - icol
      icol = icol - 1
      jcol = jcol - 1
      GO TO 30
    END IF ! (c(ndiag) /= zero)

    l = ndiag
    DO j = irow, nrow
      c(l) = zero
      l = l + j
    END DO

    60 ndiag = ndiag - irow
    irow = irow - 1
    IF (irow /= 0) GO TO 10
  END IF
END IF
RETURN

END SUBROUTINE syminv



SUBROUTINE chola(a, n, u, nullty, ifault, rmax, r)

!     ALGORITHM AS6, APPLIED STATISTICS, VOL.17, 1968, WITH
!     MODIFICATIONS BY A.J.MILLER

!     ARGUMENTS:-
!     A()    = INPUT, A +VE DEFINITE MATRIX STORED IN LOWER-TRIANGULAR
!                FORM.
!     N      = INPUT, THE ORDER OF A
!     U()    = OUTPUT, A LOWER TRIANGULAR MATRIX SUCH THAT U*U' = A.
!                A & U MAY OCCUPY THE SAME LOCATIONS.
!     NULLTY = OUTPUT, THE RANK DEFICIENCY OF A.
!     IFAULT = OUTPUT, ERROR INDICATOR
!                 = 1 IF N < 1
!                 = 2 IF A IS NOT +VE SEMI-DEFINITE
!                 = 0 OTHERWISE
!     RMAX   = OUTPUT, AN ESTIMATE OF THE RELATIVE ACCURACY OF THE
!                DIAGONAL ELEMENTS OF U.
!     R()    = OUTPUT, ARRAY CONTAINING BOUNDS ON THE RELATIVE ACCURACY
!                OF EACH DIAGONAL ELEMENT OF U.

!     LATEST REVISION - 1 April 1985

!***************************************************************************

REAL (double), INTENT(IN)   :: a(:)
INTEGER, INTENT(IN)     :: n
REAL (double), INTENT(OUT)  :: u(:)
INTEGER, INTENT(OUT)    :: nullty, ifault
REAL (double), INTENT(OUT)  :: rmax, r(:)

!     ETA SHOULD BE SET EQUAL TO THE SMALLEST +VE VALUE SUCH THAT
!     1._dp + ETA IS CALCULATED AS BEING GREATER THAN 1._dp IN THE ACCURACY
!     BEING USED.

REAL (double), PARAMETER :: eta = EPSILON(1.0_double), zero = 0.0_double
INTEGER              :: i, icol, irow, j, k, l, m
REAL (double)            :: rsq, w

ifault = 1
IF (n > 0) THEN
  ifault = 2
  nullty = 0
  rmax = eta
  r(1) = eta
  j = 1
  k = 0

!     FACTORIZE COLUMN BY COLUMN, ICOL = COLUMN NO.

  DO  icol = 1, n
    l = 0

!     IROW = ROW NUMBER WITHIN COLUMN ICOL

    DO  irow = 1, icol
      k = k + 1
      w = a(k)
      IF (irow == icol) rsq = (w*eta) ** 2
      m = j
      DO  i = 1, irow
        l = l + 1
        IF (i == irow) EXIT
        w = w - u(l) * u(m)
        IF (irow == icol) rsq = rsq + (u(l)**2*r(i)) ** 2
        m = m + 1
      END DO

      IF (irow == icol) EXIT
      IF (u(l) /= zero) THEN
        u(k) = w / u(l)
      ELSE
        u(k) = zero
        IF (ABS(w) > ABS(rmax*a(k))) GO TO 60
      END IF
    END DO

!     END OF ROW, ESTIMATE RELATIVE ACCURACY OF DIAGONAL ELEMENT.

    rsq = SQRT(rsq)
    IF (ABS(w) > 5.*rsq) THEN
      IF (w < zero) RETURN
      u(k) = SQRT(w)
      r(i) = rsq / w
      IF (r(i) > rmax) rmax = r(i)
    ELSE
      u(k) = zero
      nullty = nullty + 1
    END IF

    j = j + icol
  END DO
  ifault = zero
END IF
60 RETURN

END SUBROUTINE chola


END MODULE omMinim_mod
