! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
module ommode_mod

use omnumerics_mod
!use robustscale_mod

!******************************************
! Various algorithms for computing the mode
! using the algorithm described in the paper-
! "Robust estimators of the mode and skewness
! of continuous data", Bickel, D.R., 2002,
! Computational Statistics and Data analysis,
! Volume 39, Issue 2, pages 153-163.
! IDL code for the algorithm is given in the paper.
!
! The algorithm has been slightly ammended by CB for low pixel values 
!
!
! ommode11 - computes the modal value of a 2-d image
! ommode12 - Computes the mode when given a list of unsorted pixel
! ommode13 - Computes the mode by a jack-knife method from a -
! given number of data simulations
! 
!*******************************************************


public

interface omMode
    module procedure omMode11, omMode12, ommode13
end interface omMode


real(single), private :: minWidth = 4.01
integer, private :: minSize = 2





contains

!***************************************************
!
!
!*****************************************************
subroutine setMinWidth(newWidth)
    real(single), intent(in) :: newWidth

    minWidth = newWidth

end subroutine setMinWidth
!***************************************************
!
!
!*****************************************************
subroutine setMinSize(newSize)
    integer, intent(in) :: newSize

    minSize = newSize

end subroutine setMinSize



!********************************************************
!
! Computes the mode of a 2-d image
!
!*******************************************************
real(single) function ommode11(image, nx, ny)
    real(single), pointer :: image(:, :)
    integer(int16), intent(in) :: nx, ny

    integer(int16) :: i, j
    integer :: n
    real(single), pointer :: data(:) => null()


    data => allocateSingle(1, (nx + 1) * (ny + 1), "data")

    n = 0
    do j = 0, ny - 1

        do i = 0, nx - 1
            if (image(i, j) .ge. 0.0) then 
                n = n + 1
                data(n) = image(i, j)
            end if
        end do
    end do

    ommode11 = ommode12(n, data)



    deallocate(data)

end function ommode11
!******************************************************************************
!
!
!
!****************************************************************************
real(single) function jackknifemode(ndata, data)


    integer, intent(in) :: ndata
    real(single) :: data(:)
    integer :: i, j, k, n
    real(single), pointer, dimension(:) :: tempdata1 => null(), tempData2 => null()

    real(single), pointer :: modedata(:) => null()
    real(single) :: mode, jackMode
!    integer, save :: count = 0
    real(double) :: sum, c1, c2

!    if(count < 100) then
!        write(*, *) "jackknifemode line 76"
!    end if


    tempdata1 => allocateSingle(1, ndata, "tempData1")


  
    
!    if(count < 100) then
!        write(*, *) "jackknifemode line 86"
!    end if


  
    
    n = 0
    do i = 1, nData
        if ( .not. isNullValue(data(i)) ) then
            n = n + 1
            tempdata1(n) = data(i)
        end if
    end do
    if(n < 10) then

        jackKnifeMode = 0.0
        deallocate(tempdata1)
        return
    end if





 !   if(count < 100) then
 !       write(*, *) "jackknifemode line 111"
 !   end if





    !***************************************************
    ! Sort the data
    !
    !**************************************************
    call HeapSort( tempData1, n)
    





 !   if(count < 100) then
 !       write(*, *) "jackknifemode line 130"
 !   end if





    mode = sortedMode(n, tempData1)
     
    modedata => allocateSingle(1, n, "modeData")

    tempdata2 => allocateSingle(1, n - 1, "tempData2")

    sum = 0.0_double
    c1 = dble(n) * mode






 


   
    do i = 1, n

        k = 0
        do j = 1, i - 1    
            k = k + 1
            tempData2(k) = tempData1(j)
        end do
        do j = i + 1, n    
            k = k + 1
            tempData2(k) = tempData1(j)
        end do

        jackMode = sortedMode(n - 1, tempData2)
        sum = sum + dble(jackMode)
        c2 = dble(n - 1) * jackMode
        modeData(i) = c1 - c2
   
!     if(count < 100) then
!        write(*, *) "jackknifemode line 127 mode=", modeData(i)
!    end if

    end do

 
    jackKnifeMode = sum / dble(n)
 
!   if(count < 100) then
!        count = count + 1
!        write(*, *) "Initial mode=", mode, " jackKnife mode=", jackKnifeMode
!    end if

    deallocate(modeData, tempData1, tempData2)



end function jackknifemode


!********************************************************
!
! Computes the mode when the input data is sorted in increasing
! order
!*******************************************************
real(single) function sortedMode(ndata, data)

    integer, intent(in) :: ndata
    real(single) :: data(:)
  
    integer i, iStart, iFinish, size, n, n1, n2 
 
    real(single) :: constant, mean


   
    
 
 !   minSize = min(ndata, max(10, nData / 10))
   
   
    n = nData    
    iStart = 1
    iFinish = n
    n1 = nData
    n2 = nData
    constant = 0.5

!    if(k < 10) then
!        k = k + 1
!        write(*, *) "sortedmode line 237- the data"
!        do j = 1, n
!            write(*, *) j, ") pixel=", data(j)
!        end do
!    end if   

    do i = 1, 1000
   
        call maxInterval(data, iStart, iFinish, constant)
        size = iFinish - iStart + 1
        
!       	write(*, *) i, ") istart=", istart, " ifinish=", ifinish

	if (size <= minSize .or. size == n) then   
        
            
	    mean = sum(data(iStart : ifinish) ) / dble(size)
            sortedmode = mean 

!            if(k < 11) then
!                write(*, *) "exiting"
!                do j = istart, ifinish
!                    write(*, *) j, ") pixel=", data(j)
!                end do
!            end if
            exit
          
        end if
        
        n = size

    end do

end function sortedMode	

!********************************************************
!
! Computes the mode by a jack-knife method-
! Computes nsims data sets from the input data,
! computes the mode of each one and then 
! computes the average mode (This is supposed
! to result in a more reliable estimate of
! the mode by smoothing out bumps in the pixel
! distribution
! 
!*******************************************************
real(single) function ommode13(ndata, data, nsims)

    integer, intent(in) :: ndata
    real(single) :: data(:)
    integer, intent(in) :: nsims
    integer :: i, j, k, n     
    real(single), pointer, dimension(:) :: tempdata1 => null(), tempData2 => null()

    real(single), pointer :: modedata(:) => null()
    real(single) :: number, f1
  

    tempdata1 => allocateSingle(1, ndata, "tempData1")
 
    
    n = 0
    do i = 1, nData
        if ( .not. isNullValue(data(i)) ) then
            n = n + 1
            tempdata1(n) = data(i)
        end if
    end do
    if(n < 10) then

        ommode13 = 0.0
        deallocate(tempdata1)
        return
    end if

    !***************************************************
    ! Sort the data
    !
    !**************************************************
    call HeapSort( tempData1, n)
    


    f1 = sortedMode(n, tempData1)

    ommode13 = f1
    return



!   count = count + 1
!    write(*, *) "ommode1_mod.f90 line 312"
!    write(*, *) "Mode=", f1, " mean=", sum(tempData1)/ dble(n), " median=", sortedMedian(tempdata1, n)
!    if(count == 100) then
!    call fatal("stopped")
!    end if

    tempdata2 => allocateSingle(1, n, "tempData2")
    modedata => allocateSingle(1, nsims, "modeData")

    do i = 1, nsims

        do j = 1, n
            call random_number( number )
           
            k = int(number * real(n - 1) + 0.0001) + 1
        
            tempData2(j) = tempData1(k)
        end do
        call HeapSort( tempData2, n)
  
        modeData(i) = sortedMode(n, tempdata2)





    end do

    !********************************************************
    !
    ! Find the mean
    !
    !********************************************************
    ommode13 = sum(dble(modeData)) / dble(nsims)


    deallocate(tempdata1, tempData2, modedata)

end function ommode13
!********************************************************
!
! Computes the mode when given a list of unsorted pixel
!
!
!*******************************************************
real(single) function ommode12(ndata, data)

    integer, intent(in) :: ndata
    real(single) :: data(:)
  
    integer i, j, iStart, iFinish, size, n, n1, n2 
    integer :: count1, lCount, rCount, maxCount
    real(single) :: constant, mean, median     
    real(single), pointer :: tempdata(:) => null()
    real(single) :: xl, xu, maxXl, maxXu     

!    integer, save :: count = 0

!    if(count < 100) then
!        count = count + 1
!        write(*, *) "ommode12 line 428"
!    end if

    n = nData
    
    constant = 0.5
    tempdata => allocateSingle(1, n, "tempData")
    

    n = 0
    do i = 1, nData
        if ( .not. isNullValue(data(i)) ) then
            n = n + 1
            tempdata(n) = data(i)
        end if
    end do
    if(n < 10) then

        ommode12 = 0.0
        deallocate(tempdata)
        return
    end if


    !**************************************************
    !
    !
    !**************************************************
    call HeapSort( tempData, n)
    


    mean = sum(dble(tempData)) / dble(n)
    median = SortedMedian(tempData, n)
!    scale = max(0.5, 0.25 * Sn(tempData, n))
!    scale = 0.5

!    if(abs(mean - median) .le. scale) then
!        deallocate(tempdata)
!        ommode12 = mean
!        return
!    end if


    
    iStart = 1
    iFinish = n
    n1 = n
    n2 = n

!    write(*, *) "ommode1 line 165"
    do i = 1, 1000

!        if(count < 101) then        
!        write(*, *) i, ") istart=", istart, " ifinish=", ifinish
!        end if   

        call maxInterval(tempdata, iStart, iFinish, constant)
        size = iFinish - iStart + 1
        
       	

	if (size <= minSize .or. size == n) then   
        
            
	    mean = sum(tempdata(iStart : ifinish) ) / dble(size)
            ommode12 = mean 

!            call refineMode(n1, tempdata, mean)

!            if(count < 101) then 
!                write(*, *) "Computed mode=", ommode12, " no of iterations=", i
!                write(*, *) "The final pixels"
!                do j = iStart, iFinish
!                    write(*, *) j, ") pixel=", tempdata(j)
!                end do
!            end if

!            write(*, *) "mean=", mean, " median=", median
            exit

        end if
        
        n = size



    end do
 deallocate(tempdata)

return

!**************************************************************************
!
! EXPERIMENTAL
!
!**************************************************************************


!  write(*, *) "ommode12- intial mode=", mean


    maxCount = 0
    do i = -50, 50

        xl = real(i) + mean - 0.5
        if(xl .lt. 0.0) cycle
        if(xl .ge. tempData(n2)) cycle

        xu = real(i) + mean + 0.5
        count1 = 0
        do j = 1, n2
            if(tempData(j) .ge. xl .and. tempdata(j) .lt. xu) then
                count1 = count1 + 1
            end if 
        end do
        if(count1 .gt. maxCount) then
            maxCount = count1
            maxXl = xl
            maxXu = xu
        end if
    end do

     
    xl = maxXl - 1.0
    xu = maxXl
    lCount = 0
    do j = 1, n2
    
        if(tempData(j) .ge. xl .and. tempdata(j) .lt. xu) then
            lCount = lCount + 1
            
        endif
    end do 
  

    xl = maxXu
    xu = maxXu + 1.0
    rCount = 0
    do j = 1, n2
       
        if(tempData(j) .ge. xl .and. tempdata(j) .lt. xu) then
            rCount = rCount + 1
             
        end if
    end do 
   

    ommode12 = ommode12 + ( real(lcount - rCount) / real(lCount - 2 * count1 + rCount)) * 0.5


!   write(*, *) "Maximum bin=", maxXl, " to ", maxXu, " maxXount=", maxCount, " n=", n2, &
!    " lcount=", lcount, " rcount=", rCount, " mode=", ommode12



    deallocate(tempdata)


end function ommode12	
!**********************************************************************************
! Subroutine refinemode
!*********************************************************************************
subroutine refineMode(ndata, data, mode)
    integer, intent(in) :: ndata
    real(single), pointer :: data(:)
    real(single), intent(inout) :: mode

    real(single) :: xmin, xmax, mean, binSize, lbin, rbin, cbin
    integer :: i, j, k, jl1, ju1, jl2, ju2, jl3, ju3, count, lcount, rcount
    real(single) offset
   
!integer, save :: k9 = 0

return
    binSize = 0.5
   


    mean = max(0.0, real(floor(mode)) )

 

    write(*, *) "The data (input mode=", mode, ")"


    do i = 1, ndata
        write(*, *) i, ") f=", data(i)
    end do


              
iLoop : do i = 1, 1
        binSize = max(1.0, 0.5 * real(i) * sqrt(mean) )
      write(*, *) i, ") binsize=", binsize, " mean=", mean
        offset = 0.0
        binsize = 1.0
        do j = 1, 20
        
    
            xmin = mean + offset - binSize / 2.0

                
 
            
            xmax = mean + offset + binSize / 2.0


write(*, *) j, ") xmin=", xmin, " xmax=", xmax


            k = 1



            !***********************************************************
            ! The left bin
            !*********************************************************** 
            jl1 = LocateMinIndex( ndata, data, xmin - binsize, k)
            ju1 = LocateMaxIndex( ndata, data, xmin, jl1)

            if (ju1 <= jl1 ) then
                lCount = 0
                ju1 = 0
            else
                lCount = ju1 - jl1 + 1
            end if  

            !***********************************************************
            ! The Central bin
            !*********************************************************** 
 
            jl2 = ju1 + 1
            ju2 = LocateMaxIndex( ndata, data, xmax, jl2)

            count = ju2 - jl2 + 1
  
            !***********************************************************
            ! The Right bin
            !*********************************************************** 

            jl3 = ju2 + 1
            ju3 = LocateMaxIndex( ndata, data, xmax + binSize, jl2)
            rCount = ju3 - jl3 + 1
  

write(*, *) "xmin=", xmin, " xmax=", xmax, " jl1=", jl1, " ju1=", ju1, " jl2=", jl2, &
                  " ju2=", ju2, " jl3=", jl3, " ju3=", ju3, &
               " lcount=", lcount, " count=", count, " rcount=", rcount


            if(  count > 0 .and. count .ne. (lcount + rcount) .and. &
                (count .ge. lcount .or. count .ge. rcount) ) then
                  
           
                lbin = xmin - binSize / 2.0
                rbin = xmax + binSize / 2.0       
                cbin = mean + offset           
                write(*, *) "exiting at line 250"
                exit iloop  

            end if



            if ( lCount >= count) then
               offset = offset - 0.25        
         
               if( (mean + offset) < binsize) then
                 

                    lbin = xmin - binSize / 2.0
                    rbin = xmax + binSize / 2.0       
                    cbin = mean + offset           
                    write(*, *) "exiting at line 250"
                    exit iloop  
                end if  

          
            else
                offset = offset + 0.25
            end if
        
        
          
        end do 

write(*, *) "i=", i, " offset=", offset, " binsize=", binsize
call fatal("bug at line 258")
    end do iloop

  if( (lcount == 0 .and. rcount == 0) .or. count == 0 ) then
  write(*, *) "lcount=", lcount, " count=", count, " rcount=", rcount
  call fatal("bug")
  endif

    mean = mode    
    mode = cbin + (binSize / 2.0) * dble(lcount - rcount) / dble(lcount - 2 * count + rcount)
    

if(abs(mode - mean) > 100.0) then

write(*, *) "old mode=", mean, " new=", mode, " lcount=", lcount, &
" coint=", count, " rcount=", rcount, " cbin=", cbin, " binsize=", binsize

call fatal("bug")
endif






!if(k9 > 10) then
!call fatal("stopped")
!else
!k9 = k9 + 1
!end if

end subroutine refineMode


!**************************************************************
!
!**************************************************************

integer function LocateMinIndex(ndata, data, xmin, jStart)
    integer, intent(in) :: ndata
    real(single) :: data(:)
    real(single), intent(in) :: xmin

    integer, intent(in) :: jStart
    

    integer :: i

    do i = jStart, ndata

        if( data(i) >= xmin) then
            LocateMinIndex = i
            return 

        end if

    end do

    LocateMinIndex = jStart

end function LocateMinIndex

!**************************************************************
!
!**************************************************************

integer function LocateMaxIndex(ndata, data, xmax, jStart)
    integer, intent(in) :: ndata
    real(single) :: data(:)
    real(single), intent(in) :: xmax

    integer, intent(in) :: jStart
    

    integer :: i


    do i = jStart, ndata

        if( data(i) >= xmax) then
            LocateMaxIndex = i - 1
            return 

        end if

    end do

    LocateMaxIndex = jStart
end function LocateMaxIndex



!*********************************************************************************
!
! Finds the interval istart to ifinish that contains the maximum
! number of points such that width in pixel counts is no greater 
! than halfwidth.
! The pixels have to have been sorted into increasing order
!
!********************************************************************************* 
subroutine maxInterval(data, iStart, ifinish,  constant)
    real(single) :: data(:)
    integer, intent(inout) :: iStart, iFinish
    real(single), intent(in) :: constant
  
    integer :: i, j, maxCount, noBins, iu, count1, n, n1 
    real(single) :: width, halfWidth, wMin
    real(single) :: fPLus, d1, d2

    integer :: istart1, ifinish1, iu1
  
    istart1 = istart
    ifinish1 = ifinish

 
    width = data(iFinish) - data(iStart)
  
    if (width < minWidth) return

    halfWidth = width * constant
  


    wMin = huge(wMin)    

    maxCount = 0
    noBins = 0

    n = iFinish - iStart + 1 ! The number of points
     
   
    iu = iStart

  


    do i = iStart, iFinish - 1
   
    
    
        fPlus  = data(i) + halfWidth
        iu1 = iu
        iu = iFinish
        do j = iu1, iFinish
            if(data(j) > fPlus) then

                iu = j - 1
                exit
            end if
             
        end do
         
    	
	   
        count1 = iu - i + 1
      	 
	if (count1 == n) then ! Finished
            istart = i
            iFinish = iu
	    return
        end if
        width = data(iu) - data(i) ! The new minimum interval width
        width = max(width, minWidth)
       
       
    
        if (count1 > maxCount) then ! Maximum no of points 		 
	       
            wMin = width
                  
            noBins = 1
	    iStart1 = i
            iFinish1 = iu
 
            maxCount = count1


        else if (count1 == maxCount) then
	
	    
            if (width < wMin)then
	   
	        wMin = width   ! Smallest bin width  
              
       
                noBins = 1
                
                iStart1 = i
                iFinish1 = iu
                   
            else if(abs( width - wMin) < 0.0000001) then ! Bin widths equals minimum bin width

                noBins = noBins + 1
                iFinish1 = iu
 
            end if
	 
        end if 

        if(iu == iFinish) exit
    end do 

    if (noBins == 1) then ! Only 1 bin
  

        iStart = iStart1
        iFinish = iFinish1

        return

    end if


    !********************************************************************
    !
    ! More than one bin with the same width and count
    !
    !********************************************************************
    n1 = iFinish1 - iStart1 + 1    

    
    iStart = istart1
    iFinish = ifinish1

    if (n == n1) then

        d1 = data(istart + 1) - data(istart)
        d2 = data(ifinish) - data(ifinish - 1) 
        if ( d1 < d2 ) then
            iFinish = iFinish - 1
        elseif(d1 > d2) then
            istart = istart + 1
        else
            istart = istart + 1
            ifinish = ifinish - 1
        end if  
    end if  


end subroutine Maxinterval

end module ommode_mod



































