! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!*******************************************
! A module to compute the position and shape of
! the detected sources by moments
!******************************************
module OmParameters_mod
   
    use dal
    use OmAllocation_mod
    use errorHandling
    use omconstants_mod
    use omsourceparam_mod 

    use omnumerics_mod
    implicit none

public


interface ComputeMoments
module procedure ComputeMoments1, ComputeMoments2



end interface ComputeMoments

    
   
contains
!*************************************************************
! Subroutine ComputePointSourceParameters
! compute the source parameters- position,
! count  rates, etc
! Parameters:
! newImage - A 3-d array containing a 2-d image for each array.
! background - A 2-d background image.
! Returns: srcParams- a 1-d array of structures containing
! the parameters for each source. 
!*************************************************************
subroutine ComputePointSourceParameters(sourceNumber, nx1, ny1, image, &
                i0, j0, srcParams)


    integer(int16), intent(in) :: sourceNumber
    integer(int32), intent(in) :: nx1, ny1
  
    
    real(single), pointer   :: image(:, :)
    integer(int32), intent(in) :: i0, j0
   
    type(SourceParameters), intent(inout) :: srcParams
    
    integer (int32) :: il, iu, jl, ju
 
    real (single)   :: aValue, bValue, theta
    real (single)   :: xPos, yPos
    real (single)   ::xPosErr, yPosErr
    integer(int32) :: dx
    logical :: tooNearXAxis, tooNearYAxis 

    tooNearXAxis = .false.
    tooNearYAxis = .false.
    
    dx = 6 / binX    ! was 8
    

    il = i0 - dx
    if (il < 0)  tooNearYAxis = .true.
    il = max(0, il)
    iu = i0 + dx
    if (iu > nx1)  tooNearYAxis = .true.
    iu = min(nx1, iu)
 
    jl = j0 - dx
    if (jl < 0)  tooNearXAxis = .true.
 
    jl = max(0, jl)

    ju = j0 + dx
    if (ju > ny1)  tooNearXAxis = .true.
 
    ju = min(ny1, ju)

    srcParams%number = sourceNumber
    srcParams%iMin = il
    srcParams%iMax = iu
    srcParams%jMin = jl
    srcParams%jMax = ju
!write(*, *) " in ComputePointSourceParameter, i0=", i0, " j0=", j0
!if (srcParams%jmax .lt. 0) call fatal("bug at line 89 in ComputePointSourceParameters")  


    call ComputeMoments(sourceNumber, image, il, iu, jl, ju, &
                          xPos, yPos, xPosErr, yPosErr, &
                          aValue, bValue, theta)
  



    if (tooNearYAxis) then ! Too near the left or right-hand side of the image
        srcParams%xPos = real(i0) + 0.5
    else
!        x0 =  real(i0) + 0.5
!        if (abs(x0 - xPos) .gt. real(4 / binX)) then
!            srcParams%xPos = x0 !
!        else  
            srcParams%xPos = xPos !
!        end if                 
    end if
    if (tooNearXAxis) then ! Too near the bottom or top  of the image
        srcParams%yPos = real(j0) + 0.5
  
    else
        
!        y0 =  real(j0) + 0.5
!        if (abs(y0 - yPos) .gt. real(4 / binX)) then
!            srcParams%yPos = y0 !
!        else  
            srcParams%yPos = yPos !
!        end if                 
   
    end if              
!write(*, *) "0.5 added xsource=", srcParams%xsource, " & ",  srcParams%ysource 

    srcParams%poserr = sqrt(xPosErr * xPosErr + yPosErr * yPosErr)                

    srcParams%semiB = max(aValue, bValue)
    srcParams%semiA = min(aValue, bValue)

  
    srcParams%theta = theta * radiansToDegrees
    

 
    srcParams%noSourcePixels = 0

end subroutine ComputePointSourceParameters

!*************************************************************
! Subroutine ComputeExtendedSourceParameters
! compute the source parameters- position,
! count  rates, etc
! Parameters:
! newImage - A 3-d array containing a 2-d image for each array.
! background - A 2-d background image.
! Returns: srcParams- a 1-d array of structures containing
! the parameters for each source. 
!*************************************************************
subroutine ComputeExtendedSourceParameters(sourceNumber, nx1, ny1, image, &
                levelImage, srcParams)


    integer(int16), intent(in) :: sourceNumber
    integer(int32), intent(in) :: nx1, ny1
      
    real(single), pointer   :: image(:, :)
    integer(int16), pointer :: levelImage(:, :)
   
    type(SourceParameters), intent(inout) :: srcParams
    
    integer(int16) :: i, j
    real(single)   :: aValue, bValue, theta
    real(single)   :: xPos, yPos
    real(single)   ::xPosErr, yPosErr
    integer(int32) :: noSourcePixels
    
    i = nx1
    i = ny1    
    ! Get the source position, countrates and background count rates
    ! for each source
    ! newImage is a 3-d array (nx, ny, nUseSource)
    
    call ComputeMoments(sourceNumber, image, levelImage, int(srcParams%iMin),  &
                     int(srcParams%iMax),  int(srcParams%jMin),  int(srcParams%jMax), &
                          xPos, yPos, xPosErr, yPosErr, &
                          aValue, bValue, theta)
 
   

  
    srcParams%number = sourceNumber
 
    srcParams%xPos = xPos                
    srcParams%yPos = yPos                
    srcParams%poserr = sqrt(xPosErr * xPosErr + yPosErr * yPosErr)                
    srcParams%semiA = max(aValue, bValue)
    srcParams%semiB = min(aValue, bValue)


    srcParams%theta = theta * radiansToDegrees
  
    ! The number of pixels associated with the source
      
    noSourcePixels = 0    

    do i =  srcParams%iMin, srcParams%iMax 
        do j = srcParams%jMin, srcParams%jMax
            if (levelImage(i, j) .eq. sourceNumber) then
               
              
                noSourcePixels = noSourcePixels + 1
            end if
        end do
    end do

    ! If only a small extended source turn it into a point-source
!    if (noSourcePixels .lt. 10 / binX) then  ! was 20
!        srcParams%extensionFlag = 0
!        srcParams%noSourcePixels = 0 
!    else
        srcParams%noSourcePixels = noSourcePixels
!    end if   
end subroutine ComputeExtendedSourceParameters
!********************************************************
! subroutine ComputeMoments
! Purpose: Loops through the number of sources found for each 
! source in the crude source list (this value will normally be 1
! and hence this routine will usually be called once for each 
! of the initially detected sources)
! Each source is parameterised using moments
! Returns the x and y positions, countrates 
! and background rates for each source
! Parameters:
! 
!********************************************************


subroutine ComputeMoments1(sourceNumber, image, il, iu, jl, ju, &
                          xPos, yPos, xPosErr, yPosErr, &
                          aValue, bValue, theta)

    integer(int16), intent(in) :: sourceNumber
 
   
    real(single), pointer    :: image(:, :)          

    integer(int32), intent(in) :: il, iu, jl, ju
  
    real(single), intent(out)   :: xPos
    real(single), intent(out)   :: yPos
    real(single), intent(out)   :: xPosErr
    real(single), intent(out)   :: yPosErr
    real(single), intent(out)   :: aValue
    real(single), intent(out)   :: bValue
    real(single), intent(out)   :: theta
 
    real(single) :: x2
    real(single) :: y2
    real(single) :: xy
    real(single) :: val1, val2, diff
        
    integer(int32) :: i, j, i1, j1
   
    real(double) :: imageVal
   
    real(double) :: xErrTot, yErrTot
    real(double) :: xposD, yposD, x2D, y2D, xyD
    real(double) :: totalImage, totalImage2    
   

    totalImage = 0.0_double
    totalImage2 = 0.0_double
    
    xposD = 0.0_double
    yposD = 0.0_double
    x2D = 0.0_double
    y2D = 0.0_double
    xyD = 0.0_double


    !********************************************************************
    ! Compute the x and y centroid positions (xpos & ypos)
    !********************************************************************

    do i = il, iu
        i1 = i - il + 1
        do j = jl, ju
            imageVal = dble(image(i, j))
           
            if(imageVal .ne. pixelIgnoreFlag) then
                xposD = xposD + imageVal * dble(i1)   
                yposD = yposD + imageVal * dble(j - jl + 1)   
                totalImage2 = totalImage2 + imageVal * imageVal
                totalImage = totalImage + imageVal
                  
            end if
        end do
    end do

      
     
    xErrTot = 0.0_double
    yErrTot = 0.0_double
    xposD = xposD / totalImage
    yposD = yposD / totalImage

    xpos = real(xposD)  
    ypos = real(yposD) 

!write(*, *) "line 298 xposd=", xposD, " yposD=", yposD
!write(*, *) "totalimage=", totalImage

!write(*, *) "totalimage2=", totalImage2
    !*******************************************************
    ! Compute the postion errors
    !*******************************************************

    do i = il, iu
        i1 = i - il + 1
        do j = jl, ju
            j1 = j - jl + 1 
            imageVal = dble(image(i, j)) 
            if(imageVal .ne. pixelIgnoreFlag) then

                !*******************************************************
                ! Assume that the error is dominated by the total counts
                !*******************************************************
                xErrTot = xErrTot + imageVal * dble((j1 - xposD) ** 2)
                yErrTot = yErrTot + imageVal * dble((i1 - yposD) ** 2)
               
                x2D = x2D + imageVal * dble(i1 * i1)
                y2D = y2D + imageVal * dble(j1 * j1)
                xyD = xyD + imageVal * dble(i1 * j1)
            end if
        end do
    end do

    xPosErr = sqrt(xErrTot / totalImage2)   ! Source pixel x-position error
    yPosErr = sqrt(yErrTot / totalImage2)   ! Source pixel y-position error

    x2 = real(x2D / totalImage - xposD * xposD)
    y2 = real(y2D / totalImage - yposD * yposD)
    xy = real(xyD / totalImage - xposD * yposD)





    xPos = xPos + real(il)   ! PIXCOORD1 SYSTEM (mid point of first pixel = (1, 1)
    yPos = yPos + real(jl)

    ! Now convert these into major/minor values and theta
    ! factor of 2 somewhere?
    
    val1 = (x2 + y2) / 2.0
    val2 = sqrt(((x2 - y2) / 2.0) ** 2 + xy ** 2)
    aValue = sqrt(val1 + val2)

        
    if(val1 .gt. val2) then
        bValue = sqrt(val1 - val2)
    else
        diff = abs(val1 - val2)

        if (diff .gt. 1.0) then ! NB original JPDM criterion was 0
            
            ! There is a problem with the source parameterisation
          
            write(errstr,'("Warning: Problem with parameterisation of source", i6)') sourceNumber
  
            call message(AppMsg, Verbose)      ! Verbose was 2
            
        else

            bValue = aValue
        end if
    end if 

          
    ! The position angle of the source
    theta = Angle(x2, y2, xy)
            
  

end subroutine ComputeMoments1

!********************************************************
! subroutine ComputeMoments
! Purpose: Loops through the number of sources found for each 
! source in the crude source list (this value will normally be 1
! and hence this routine will usually be called once for each 
! of the initially detected sources)
! Each source is parameterised using moments
! Returns the x and y positions, countrates 
! and background rates for each source
! Parameters:
! image - A 3-d array- one 2-d image for each source
!********************************************************

subroutine ComputeMoments2(sourceNumber, image, levelImage, il, iu, jl, ju, &
                          xPos, yPos, xPosErr, yPosErr, &
                          aValue, bValue, theta)


    integer(int16), intent(in) :: sourceNumber
 
   
    real(single), pointer    :: image(:, :)          
    integer(int16), pointer :: levelImage(:, :)

    integer(int32), intent(in) :: il, iu, jl, ju
  
    real(single), intent(out)   :: xPos
    real(single), intent(out)   :: yPos
    real(single), intent(out)   :: xPosErr
    real(single), intent(out)   :: yPosErr
    real(single), intent(out)   :: aValue
    real(single), intent(out)   :: bValue
    real(single), intent(out)   :: theta
 
    real(single) :: x2
    real(single) :: y2
    real(single) :: xy
    real(single) :: val1, val2, diff
        
    integer(int32) :: i, j, i1, j1
   
    real(double) :: imageVal
   
    real(double) :: xErrTot, yErrTot
    real(double) :: xposD, yposD, x2D, y2D, xyD
    real(double) :: totalImage, totalImage2    
   

    totalImage = 0.0_double
    totalImage2 = 0.0_double
    
    xposD = 0.0_double
    yposD = 0.0_double
    x2D = 0.0_double
    y2D = 0.0_double
    xyD = 0.0_double


    !********************************************************************
    ! Compute the x and y centroid positions (xpos & ypos)
    !********************************************************************

    do i = il, iu
        i1 = i - il + 1
        do j = jl, ju
            imageVal = dble(image(i, j))
           
            if(imageVal .ne. pixelIgnoreFlag .and. levelImage(i, j) .eq. sourceNumber) then
                xposD = xposD + imageVal * dble(i1)   
                yposD = yposD + imageVal * dble(j - jl + 1)   
                totalImage2 = totalImage2 + imageVal * imageVal
                totalImage = totalImage + imageVal
                  
            end if
        end do
    end do
          
    xErrTot = 0.0_double
    yErrTot = 0.0_double
    xposD = xposD / totalImage
    yposD = yposD / totalImage

    xpos = real(xposD)  
    ypos = real(yposD) 

!write(*, *) "xposd=", xposD, " yposD=", yposD
!write(*, *) "totalimage2=", totalImage2
    !*******************************************************
    ! Compute the postion errors
    !*******************************************************

    do i = il, iu
        i1 = i - il + 1
        do j = jl, ju
            j1 = j - jl + 1 
            imageVal = dble(image(i, j)) 
            if(imageVal .ne. pixelIgnoreFlag .and. levelImage(i, j) .eq. sourceNumber) then

                !*******************************************************
                ! Assume that the error is dominated by the total counts
                !*******************************************************
                xErrTot = xErrTot + imageVal * dble((j1 - xposD) ** 2)
                yErrTot = yErrTot + imageVal * dble((i1 - yposD) ** 2)
               
                x2D = x2D + imageVal * dble(i1 * i1)
                y2D = y2D + imageVal * dble(j1 * j1)
                xyD = xyD + imageVal * dble(i1 * j1)
            end if
        end do
    end do

    xPosErr = sqrt(xErrTot / totalImage2)   ! Source pixel x-position error
    yPosErr = sqrt(yErrTot / totalImage2)   ! Source pixel y-position error

    x2 = real(x2D / totalImage - xposD * xposD)
    y2 = real(y2D / totalImage - yposD * yposD)
    xy = real(xyD / totalImage - xposD * yposD)





    xPos = xPos + real(il)   ! PIXCOORD1 SYSTEM (mid point of first pixel = (1, 1)
    yPos = yPos + real(jl)

    ! Now convert these into major/minor values and theta
    ! factor of 2 somewhere?
    
    val1 = (x2 + y2) / 2.0
    val2 = sqrt(((x2 - y2) / 2.0) ** 2 + xy ** 2)
    aValue = sqrt(val1 + val2)

        
    if(val1 .gt. val2) then
        bValue = sqrt(val1 - val2)
    else
        diff = abs(val1 - val2)

        if (diff .gt. 1.0) then ! NB original JPDM criterion was 0
            
            ! There is a problem with the source parameterisation
          
            write(errstr,'("Warning: Problem with parameterisation of source", i6)') sourceNumber
  
            call message(AppMsg, Verbose)      ! Verbose was 2
            
        else

            bValue = aValue
        end if
    end if 

          
    ! The position angle of the source
    theta = Angle(x2, y2, xy)
            
end subroutine ComputeMoments2


end module OmParameters_mod



















