! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
! Module: omdetect_mod
!
! Authors:  Tim Harries - MSSL (orig) 
!           almost completely rewritten by J. Mittaz MSSL
!          
!*********************************************************************************** 
module OmSmoothImage_mod

    use dal
    use cal
    use param  
    use sasdal
 
    use getFilterID_mod
    use errorHandling
    use omallocation_mod
    use omskybackground_mod
    use splinefit_mod 
    use ompsf_mod
    use Matrix_mod

use omutility_mod
    implicit none 
 
 

interface SmoothImage
module procedure SmoothImage1, SmoothImage2, &
        SmoothImage4, KuwaharaFilter, &
        smoothimagesavgol


end interface SmoothImage


interface lpdFilter
    module procedure LPDFilter1, lpdfilter2, lpdfilter3


end interface lpdFilter





   

PUBLIC :: Savgol
public :: SmoothImageSavgol
CONTAINS

!*********************************************
! 
!********************************************
subroutine Savgol(np, c, nl, nr, ld, m)

    integer, intent(in) :: np
    real, pointer :: c(:)
    integer, intent(in) :: nl, nr, ld, m  

    integer :: imj, ipj, k, mm, l
    integer, allocatable :: index(:)
    real :: d, fac, sum
    real, allocatable :: a(:, :), b (:)
    
    if (np < nl + nr + 1 .or. nl < 0 .or. nr < 0 .or. &
        ld > m .or. nl + nr < m) then
        write(*, *) "Bad arguments in savgol"
    end if


    allocate(index(1 : m + 1))
    allocate(a(1 : m + 1, 1 : m + 1))
    allocate(b(1 : m + 1))
  
    !**************************************************************
    ! Set up the normal equations of the desired least-squares fit
    !**************************************************************    

    do ipj = 0, m * 2
        if (ipj .gt. 0) then 
            sum = 0.0
        else
            sum = 1.0
        end if

        do k = 1, nr
            sum = sum + (real(k) ** ipj)
        end do

        do k = 1, nl
            sum = sum + (real(-k) ** ipj)
        end do

        mm = min(ipj, 2 * m - ipj)

        do imj = -mm, mm, 2
            a(1 + (ipj + imj) / 2, 1 + (ipj - imj) / 2) = sum
        end do

    end do


    !*****************************************
    ! Solve the equations by LU decomposition
    !*****************************************   
    call Ludcmp(m + 1, a, index, d) 
        

    b = 0.0
    b(ld + 1) = 1.0

    !**************************************************
    ! Right-hand side vector is unit vector, depending
    ! on which derivative we want
    !**************************************************
    call Lubksb(m + 1, a, index, b)

    c = 0.0
    l = 0    
    do k = -nl, nr
        sum = b(1)
        fac = 1.0
        do mm = 1, m
            fac = fac * real(k)
            sum = sum + b(mm + 1) * fac 
        end do
       

       c(l) = sum
       l = l + 1
    end do


   

    deallocate(b, a, index)
end subroutine Savgol 

!********************************************************
! subroutine smooth array
!*******************************************************
subroutine SmoothArray(arrayPtr, nl, nr, ld, m)
    real(single) :: arrayPtr(:)
    integer, intent(in) :: nl, nr, ld, m
 
    real(single), pointer :: c(:) => null()
    integer :: np, i,il, iu, nmax
    real(single), pointer :: tempArrayPtr(:) => null()

    il = lbound(arrayPtr, 1)
    iu = ubound(arrayPtr, 1)
    nmax = size(arrayPtr)
    tempArrayPtr => allocateSingle(il, iu, "tempArrayPtr") 
    c => allocateSingle(0 , nl + nr + 1, "c in smoothimage (savgol)") 
 


!write(*, *) "Line 118"

    do i = il, nl + 1
        np = min(i - il + nr + 1, nmax)
!write(*, *) "i=", i, " np=", np, " il=", il 
        call Savgol(np, c, i - il, nr, ld, min(np - 1, m))
        tempArrayPtr(i) = SmoothArrayValue(arrayPtr, i, c, i - il, nr)

    end do

!write(*, *) "Line 127"


    np = nl + nr + 1
    call Savgol(np, c, nl, nr, ld, m)
    

    do i = il + nl, iu - nr  
        tempArrayPtr(i) = SmoothArrayValue(arrayPtr, i, c, nl, nr)
    end do
!write(*, *) "Line 137"




    do i = iu - nr + 1, iu
        np = nl + iu - i + 1 
        call Savgol(np, c, nl, iu - i, ld, min(np - 1, m))
        tempArrayPtr(i) = SmoothArrayValue(arrayPtr, i, c, nl, iu - i)

    end do
 !write(*, *) "Line 148"


    arrayPtr = tempArrayPtr
    deallocate(tempArrayPtr, c)
    nullify(tempArrayPtr, c)

end subroutine SmoothArray

!********************************************************
! Savitzky-Golay 2-d smoothing
! dx is the number of pixels either side of the
! current pixel to be used in the smoothing
! m is the order of the polynomial
!
!*******************************************************
subroutine SmoothImageSavgol(image, dx, ld, m)
    real(single), pointer :: image(:, :)
    integer, intent(in) :: dx, ld, m
    
    real(single), pointer :: c(:) => null()
    integer :: np, i, j, nx, ny, nl, nr
    real(single), pointer :: tempImage(:, :) => null()

    nx = ubound(image, 1)
    ny = ubound(image, 2)

write(*, *) "ld=", ld, " m=", m
    tempImage => allocate2DSingle(0, nx, 0, ny, "tempImage") 
    nl = 4 * (dx + 1)
    nr = nl
    np = 2 * nr + 1

    ! Compute the coefficients
    c => allocateSingle( 0, np, "c in smoothimage (savgol)") 
    c(0) = -0.0743
    c(1) =  0.0114
    c(2) =  0.0400
    c(3) =  0.0114
    c(4) = -0.0743

    c(5) =  0.0114
    c(6) =  0.0971
    c(7) =  0.1257
    c(8) =  0.0971
    c(9) =  0.0114

    c(10) =  0.0400
    c(11) =  0.1257
    c(12) =  0.1543
    c(13) =  0.1257
    c(14) =  0.0400

    c(15) =  0.0114
    c(16) =  0.0971
    c(17) =  0.1257
    c(18) =  0.0971
    c(19) =  0.0114

    c(20) = -0.0743
    c(21) =  0.0114
    c(22) =  0.0400
    c(23) =  0.0114
    c(24) = -0.0743










!    call Savgol(np, c, nl, nr, ld, m)

    tempImage = image
    
    do j = 0, ny
        do i = 0, nx
            image(i, j) = SmoothPixel( tempimage, i, j, c, dx )

        end do
    end do
    deallocate(c)
    nullify(c)
  
    deallocate(tempImage)

end subroutine SmoothImageSavgol


!******************************************************
!
!******************************************************
real(single) function SmoothPixel(image, i0, j0, c, dx)
    real(single), pointer :: image(:, :)
    integer, intent(in) :: i0, j0, dx
    real(single), pointer :: c(:)
    integer :: i, j, k, l, nx, ny
    
    real(double) :: g, sum

    nx = ubound(image, 1)
    ny = ubound(image, 2)

    g = 0.0_double
    sum = 0.0_double
    l = 0
    k = -1

    do j = j0 + dx, j0 - dx, -1
     
        do i = i0 - dx, i0 + dx
            k = k + 1
            if (j .lt. 0) cycle
            if (j .gt. ny) cycle 
    
            if (i .lt. 0) cycle
            if (i .gt. nx) cycle          
            if (image(i, j) .ge. 0.0) then
            
                g = g + dble(c(k)) * dble(image(i, j))
                l = l + 1
                sum = sum + dble(c(k))
            end if
        end do  

    end do


    if (l .gt. 0) then
        SmoothPixel = g / sum
    else
        SmoothPixel = image(i0, j0)
    end if
  
end function SmoothPixel    




!******************************************************
!
!******************************************************
real(single) function SmoothArrayValue(arrayPtr, i0, c, nl, nr)
    real(single) :: arrayPtr(:)
    integer, intent(in) :: i0, nl, nr
    real(single), pointer :: c(:)
    integer :: k, l
    real(double) :: g

    l = 0

    do k = -nl, nr
        l = l + 1
        g = g + c(l) * arrayPtr(i0 + k)
               
    end do

    SmoothArrayValue = g 

end function SmoothArrayValue    


!******************************************************
!
!******************************************************
real(single) function SmoothX(image, i0, j0, c, nl, nr)
    real(single), pointer :: image(:, :)
    integer, intent(in) :: i0, j0, nl, nr
    real(single), pointer :: c(:)
    integer :: k, l
    real(double) :: g, sum


    g = 0.0_double
    sum = 0.0_double
 
    l = 0
    do k = -nl, nr
         
        if (image(i0 + k, j0) .ge. 0.0) then
            
            g = g + dble(c(l)) * dble(image(i0 + k, j0))
            l = l + 1
            sum = sum +  dble(c(l))
        end if
  

    end do
    if (l .gt. 0) then
        SmoothX = g / sum
    else
        SmoothX = image(i0, j0)
    end if
  
end function SmoothX    


!******************************************************
!
!******************************************************
real(single) function SmoothX2(image, maskimage, i0, j0, c, nl, nr)
    real(single), pointer :: image(:, :)
    logical, pointer :: maskimage(:, :)

    integer, intent(in) :: i0, j0, nl, nr
    real(single), pointer :: c(:)
    integer :: k, l
    real(double) :: g, sumWeight


    if (maskImage(i0, j0)) then
       SmoothX2 = image(i0, j0)
       return
    end if



    g = 0.0_double
    sumWeight = 0.0_double
    l = 0
    do k = -nl, nr
         
        if (maskimage(i0 + k, j0)) cycle
            
        g = g + dble(c(l)) * dble(image(i0 + k, j0))
        sumWeight = sumWeight + dble(c(l)) 
        l = l + 1
    end do
    if (l .gt. 2) then
        SmoothX2 = g / sumWeight
    else
        SmoothX2 = image(i0, j0)
    end if
end function SmoothX2    

!******************************************************
!
!******************************************************
real(single) function SmoothY(image, i0, j0, c, nl, nr)
    real(single), pointer :: image(:, :)

    integer, intent(in) :: i0, j0, nl, nr
    real(single), pointer :: c(:)
    integer :: k, l
    real(double) :: g, sum

    g = 0.0_double
    sum = 0.0_double
    l = 0

    do k = -nl, nr
        if (image(i0, j0 + k) .ge. 0.0) then
  
       
            g = g + dble(c(l)) * dble(image(i0, j0 + k))
            l = l + 1
            sum = sum + dble(c(l))
        end if
    end do
    if (l .gt. 0) then
        SmoothY = g / sum
    else
        SmoothY = image(i0, j0)
    end if
  


end function SmoothY    


!******************************************************
!
!******************************************************
real(single) function SmoothY2(image, maskiMage, i0, j0, c, nl, nr)
    real(single), pointer :: image(:, :)
    logical, pointer :: maskimage(:, :)

    integer, intent(in) :: i0, j0, nl, nr
    real(single), pointer :: c(:)
    integer :: k, l
    real(double) :: g, sumWeight


    if (maskImage(i0, j0)) then
       SmoothY2 = image(i0, j0)
       return
    end if



    g = 0.0_double
    sumWeight = 0.0_double

    l = 0
    do k = -nl, nr
        if (maskimage(i0, j0 + k)) cycle
            
    
       
        g = g + dble(c(l)) * dble(image(i0, j0 + k))
        sumWeight = sumWeight + dble(c(l)) 
  
        l = l + 1
    end do
   
    if (l .gt. 2) then
        SmoothY2 = g / sumWeight
    else
        SmoothY2 = image(i0, j0)
    end if
end function SmoothY2    


!************************************************
!
!***********************************************
subroutine LPDFilter1(image)
    real(single), pointer :: image(:, :)
    real(single), pointer :: tempImage(:, :) => null()

    integer :: nx, ny, i, j
  
    real(single) :: weight0, weight1, weight2

    real(double) :: sum, sumWeight
   
    nx = ubound(image, 1)
    ny = ubound(image, 2)
    
    

    weight0 = 0.25_double
    weight1 = 1.0_double / 8.0_double
    weight2 = 1.0_double / 16.0_double
  

    tempImage => Allocate2DSingle(0, nx, 0, ny, "tempImage")

    tempImage = image

    do i = 0, nx
        do j = 0, ny
            if( .not. isNullValue( image(i, j) ) ) then
                sum = image(i, j) * weight0
                sumWeight = weight0
            else
                sum = 0.0_double
                sumWeight = 0.0_double 
            end if
     
            if (i .gt. 0) then ! 
               

                if( .not. isNullValue( image(i - 1, j) ) ) then
                    sum = sum + image(i - 1, j) * weight1 ! Adjacent left pixel
                    sumWeight = sumWeight + weight1
                end if
                if (j .gt. 0) then
                   
                    if( .not. isNullValue( image(i - 1, j - 1) ) ) then
       
                        sum = sum + image(i - 1, j - 1) * weight2 ! Adjacent bottom-left pixel
                        sumWeight = sumWeight + weight2
                    end if
                end if        
                if (j .lt. ny) then
                   
                    if( .not. isNullValue( image(i - 1, j + 1) ) ) then
                        sum = sum + image(i - 1, j + 1) * weight2 ! Adjacent top-left pixel
                        sumWeight = sumWeight + weight2
                    end if
                end if        

            end if

            if (j .gt. 0) then
      
                if( .not. isNullValue( image(i, j - 1) ) ) then
       
                    sum = sum + image(i, j - 1) * weight1 ! Adjacent bottom pixel
                    sumWeight = sumWeight + weight1
                end if
            end if        
            if (j .lt. ny) then
              
                if( .not. isNullValue( image(i, j + 1) ) ) then
       
                    sum = sum + image(i, j + 1) * weight1 ! Adjacent top pixel
                    sumWeight = sumWeight + weight1
                end if
            end if        

            if (i .lt. nx) then ! 
                if( .not. isNullValue( image(i + 1, j) ) ) then
                 
                    sum = sum + image(i + 1, j) * weight1 ! Adjacent right pixel
                    sumWeight = sumWeight + weight1
                end if
                if (j .gt. 0) then
                    if( .not. isNullValue( image(i + 1, j - 1) ) ) then
       
                        sum = sum + image(i + 1, j - 1) * weight2 ! Adjacent bottom-right pixel
                        sumWeight = sumWeight + weight2
                    end if 
                end if        
                if (j .lt. ny) then
                  
                    if( .not. isNullValue( image(i + 1, j + 1) ) ) then
       
                        sum = sum + image(i + 1, j + 1) * weight2 ! Adjacent top-right pixel
                        sumWeight = sumWeight + weight2
                    end if
                end if        

            end if
            if(sumweight .gt. 0.0_double) then
                tempImage(i, j) = sum / sumWeight
            end if
        end do
    end do
    
    image = tempImage
    deallocate(tempImage)
    nullify(tempImage)
 

end subroutine LPDFilter1

!************************************************
!
!***********************************************
subroutine LPDFilter2(image, qualityImage)
    real(single), pointer :: image(:, :)
   
    logical(bool), pointer :: qualityImage(:, :)
    real(single), pointer :: tempImage(:, :) => null()

    integer :: nx, ny, i, j
  
    real(single) :: weight0, weight1, weight2

    real(double) :: sum, sumWeight
  
    nx = ubound(image, 1)
    ny = ubound(image, 2)
    
    

    weight0 = 0.25_double
    weight1 = 1.0_double / 8.0_double
    weight2 = 1.0_double / 16.0_double
  

    tempImage => Allocate2DSingle(0, nx, 0, ny, "tempImage")

    tempImage = image

    do i = 0, nx
        do j = 0, ny
             
            if ( .not. qualityImage(i, j) ) then 
           
                sum = image(i, j) * weight0
                sumWeight = weight0
            else
                sum = 0.0_double
                sumWeight = 0.0_double
            end if 

            if (i .gt. 0) then ! 
               

                if ( .not. qualityImage(i - 1, j) ) then 
           

                    sum = sum + image(i - 1, j) * weight1 ! Adjacent left pixel
                    sumWeight = sumWeight + weight1
               
                end if


                if (j .gt. 0 ) then
                    if( .not. qualityImage(i - 1, j - 1)   ) then
                   
                        sum = sum + image(i - 1, j - 1) * weight2 ! Adjacent bottom-left pixel
                        sumWeight = sumWeight + weight2
                    end if
                end if        
                if (j .lt. ny ) then
                    if (.not. qualityImage(i - 1, j + 1)  ) then
                   
                        sum = sum + image(i - 1, j + 1) * weight2 ! Adjacent top-left pixel
                        sumWeight = sumWeight + weight2
                   
                    end if
                end if        

            end if

            if (j .gt. 0) then
                if(.not. qualityImage(i, j - 1)  ) then
      
                    sum = sum + image(i, j - 1) * weight1 ! Adjacent bottom pixel
                    sumWeight = sumWeight + weight1
               
                end if   
            end if        
            if (j .lt. ny ) then
                if( .not. qualityImage(i, j + 1)  ) then
              
                    sum = sum + image(i, j + 1) * weight1 ! Adjacent top pixel
                    sumWeight = sumWeight + weight1
              
                end if   
            end if        

            if (i .lt. nx ) then ! 
                if ( .not. qualityImage(i + 1, j) ) then 
           
            
                    sum = sum + image(i + 1, j) * weight1 ! Adjacent right pixel
                    sumWeight = sumWeight + weight1
                 
                end if 
                if (j .gt. 0) then
                    if ( .not. qualityImage(i + 1, j - 1) ) then 
                      
                        sum = sum + image(i + 1, j - 1) * weight2 ! Adjacent bottom-right pixel
                        sumWeight = sumWeight + weight2
                  
                    end if
                end if        
                if (j .lt. ny) then
                    if ( .not. qualityImage(i + 1, j + 1) ) then 
           
                        sum = sum + image(i + 1, j + 1) * weight2 ! Adjacent top-right pixel
                        sumWeight = sumWeight + weight2
                    
                    end if
                end if        

            end if
            if(sumWeight .ge. 0.0_double) then
                tempImage(i, j) = sum / sumWeight
            end if 
        end do
    end do
    
    image = tempImage
    deallocate(tempImage)
    nullify(tempImage)
 

end subroutine LPDFilter2


!************************************************
!
!***********************************************
subroutine LPDFilter3(image)
    real(double), pointer :: image(:, :)
    real(double), pointer :: tempImage(:, :) => null()

    integer :: nx, ny, i, j
  
    real(double) :: weight0, weight1, weight2

    real(double) :: sum, sumWeight
   
    nx = ubound(image, 1)
    ny = ubound(image, 2)
    
    

    weight0 = 0.25_double
    weight1 = 1.0_double / 8.0_double
    weight2 = 1.0_double / 16.0_double
  

    tempImage => Allocate2Ddouble(0, nx, 0, ny, "tempImage")

    tempImage = image

    do i = 0, nx
        do j = 0, ny
            if( .not. isNullValue( image(i, j) ) ) then
                sum = image(i, j) * weight0
                sumWeight = weight0
            else
                sum = 0.0_double
                sumWeight = 0.0_double 
            end if
     
            if (i .gt. 0) then ! 
               

                if( .not. isNullValue( image(i - 1, j) ) ) then
                    sum = sum + image(i - 1, j) * weight1 ! Adjacent left pixel
                    sumWeight = sumWeight + weight1
                end if
                if (j .gt. 0) then
                   
                    if( .not. isNullValue( image(i - 1, j - 1) ) ) then
       
                        sum = sum + image(i - 1, j - 1) * weight2 ! Adjacent bottom-left pixel
                        sumWeight = sumWeight + weight2
                    end if
                end if        
                if (j .lt. ny) then
                   
                    if( .not. isNullValue( image(i - 1, j + 1) ) ) then
                        sum = sum + image(i - 1, j + 1) * weight2 ! Adjacent top-left pixel
                        sumWeight = sumWeight + weight2
                    end if
                end if        

            end if

            if (j .gt. 0) then
      
                if( .not. isNullValue( image(i, j - 1) ) ) then
       
                    sum = sum + image(i, j - 1) * weight1 ! Adjacent bottom pixel
                    sumWeight = sumWeight + weight1
                end if
            end if        
            if (j .lt. ny) then
              
                if( .not. isNullValue( image(i, j + 1) ) ) then
       
                    sum = sum + image(i, j + 1) * weight1 ! Adjacent top pixel
                    sumWeight = sumWeight + weight1
                end if
            end if        

            if (i .lt. nx) then ! 
                if( .not. isNullValue( image(i + 1, j) ) ) then
                 
                    sum = sum + image(i + 1, j) * weight1 ! Adjacent right pixel
                    sumWeight = sumWeight + weight1
                end if
                if (j .gt. 0) then
                    if( .not. isNullValue( image(i + 1, j - 1) ) ) then
       
                        sum = sum + image(i + 1, j - 1) * weight2 ! Adjacent bottom-right pixel
                        sumWeight = sumWeight + weight2
                    end if 
                end if        
                if (j .lt. ny) then
                  
                    if( .not. isNullValue( image(i + 1, j + 1) ) ) then
       
                        sum = sum + image(i + 1, j + 1) * weight2 ! Adjacent top-right pixel
                        sumWeight = sumWeight + weight2
                    end if
                end if        

            end if
            if(sumweight .gt. 0.0_double) then
                tempImage(i, j) = sum / sumWeight
            end if
        end do
    end do
    
    image = tempImage
    deallocate(tempImage)
    nullify(tempImage)
 

end subroutine LPDFilter3




!********************************************************
! subroutine smoothIMage
! Routine to smooth image 
!
!********************************************************
subroutine SmoothImage1(image, nx, ny, XsmoothSize, YsmoothSize)
 
    integer(int32), intent(in)    :: nx, ny
    real(single), pointer      :: image(:, :)
    

    real(single), pointer   :: tempImage(:, :) => null()
    integer(int32), intent(in)    :: XSmoothSize, ySmoothSize

    ! internal variables
    integer(int32)                :: height
    integer(int32)                :: i

    ! For top hat
    real(single), pointer         :: scan(:) => null()
    integer(int32), pointer       :: nscan(:) => null()
    integer(int32)                :: allocErr


    tempImage => Allocate2DSingle(0, nx - 1, 0 , ny - 1, "tempImage")
     

    !*************************
    ! for gaussian smooth 
    !************************
    height = YSmoothSize
    
    scan => AllocateSingle(-height, height, "scan (smoothImage)")
    nscan => AllocateInt32(-height, height, "nscan (smoothImage)")
   

    !**************************************************
    ! smoothTopHat deals with one X position at a time
    !**************************************************
    do i = 0, nx - 1
        call MovingBox(image, tempImage, nx, ny, XSmoothSize, YSmoothSize, i, scan, nscan, 1)
    end do
  
    deallocate(scan, stat = allocErr)
    call CheckDeallocation(allocErr, "scan")

    deallocate(nscan, stat = allocErr)
    call CheckDeallocation(allocErr, "nscan")
    image = tempImage
    deallocate(tempImage)

end subroutine SmoothImage1

!******************************************************************************
! subroutine MovingBox
! Top hat smoothing routine - does a strip in Y direction (best array access)
! 12.12.2001 - Modified to use the level image (ie to ignore non-background
! pixels)
!******************************************************************************
subroutine MovingBox(image, backgroundTot, nx, ny, width, height, startX, scan, &
                     nscan, type)

    integer(int32), intent(in)  :: nx, ny
    real(single), pointer    :: image(:, :)

   
    real(single), pointer :: backgroundTot(:, :)


    integer(int32), intent(in) :: width, height     ! Size of smoothing area
    integer(int32), intent(in) :: startX           ! starting point
    real(single), pointer :: scan(:)
    integer(int32), pointer :: nscan(:)

    integer(int32), intent(in) :: type
    !***********************
    ! internal variables
    ! for top hat smooth
    !***********************
    integer(int32)            :: i, i1, j ,k
    real(single)               :: running_total
    integer(int32)            :: nrunning_total
    integer(int32)            :: minXpos, maxXpos
    integer(int32)            :: minYpos, maxYpos
    integer(int32)            :: clearStart, clearEnd
    integer(int32)            :: clearEndPlus1

    real(single)              :: scanval


    !************************************
    ! initialise arrays for this strip
    !************************************
    scan = 0
    nscan = 0

    ! X min and max's are fixed
     
    minXpos = MAX(startX - width, 0)
    maxXpos = MIN(startX + width, nx - 1)
 
    !**************************************************************
    ! Get start and end of clear (ie. don't need MIN/MAX checks )
    !**************************************************************
    clearStart = MIN(height + 1, ny)

    if(height .gt. ny) then
        clearEnd = -1
    else
        clearEnd = ny - height - 1
    end if

    !******************************************
    ! At start need to build up running total  
    ! First one
    !******************************************
    minYpos = 0
    maxYpos = MIN(height, ny - 1)


    running_total= 0.0
    nrunning_total= 0

    do j = minYpos, maxYpos     
        scanval = 0.0
        nScan(j) = 0
        do k = minXpos, maxXpos 
            if (image(k, j) .ge. 0.0) then            
                scanval = scanval + image(k, j)
                nScan(j) = nscan(j) + 1
            end if
        end do
        scan(j) = scanval
        
        running_total = running_total + scanval
        nrunning_total = nrunning_total + nscan(j)

    end do
  
    if(type .eq. 1) then
        if (nrunning_total .gt. 0) then   
            backgroundTot(startX, 0) = running_total / nrunning_total
        else
            backgroundTot(startX, 0) = 0.0
        end if
    else
        backgroundTot(startX, 0) = running_total
    end if

  
    do i = 1, clearStart - 1
     
        ! Shift down arrays
     
        do j = minYpos,  maxYpos
            scan(j - 1) = scan(j)
            nscan(j - 1) = nscan(j)
        end do
     
        ! reset min/max values
     
        minYpos = MAX(i - height, 0)
        maxYpos = MIN(i + height, ny - 1)

        !***********************
        ! Add in new entry
        !***********************     
        scanval = 0.0
        i1 = maxYpos - i
        nscan( i1 ) = 0
        
        do k = minXpos, maxXpos 
            if (image(k, maxYpos) .ge. 0.0) then
                nscan( i1 ) = nscan( i1 ) + 1  
                scanval = scanval + image(k, maxYpos)
            end if
        end do
        minYpos = minYpos - i
        maxYpos = i1
        scan(maxYpos) = scanval
      
        running_total = running_total + scanval
        nrunning_total = nrunning_total + nscan(maxYpos)
 
        if(type .eq. 1) then
            if (nrunning_total .gt. 0) then
                backgroundTot(startX, i) = running_total / nrunning_total
            else
                backgroundTot(startX, i) = 0.0
            end if
        else
            backgroundTot(startX, i) = running_total
        end if
    
    end do
 

    !********************************
    ! Now do the rest (not MIN/MAX)
    !********************************
    if(clearEnd .eq. -1) return
  

    ! Make sure we can hold whole smoothing area within data

    if(clearStart .lt. clearEnd) then
        do i = clearStart, clearEnd
     
            ! Shift down arrays and remove bottom line

            running_total = running_total - scan(-height)
            nrunning_total = nrunning_total - nscan(-height)
            do j = -height + 1, height
                scan(j - 1) = scan(j)
                nscan(j - 1) = nscan(j)
            end do
            !*******************
            ! Add in new entry
            !******************
        
            maxYpos = i + height
            scanval = 0.0
            nscan(height) = 0
            do k = minXpos, maxXpos 
                if (image(k, maxYpos) .ge. 0.0) then            
                    nscan(height) = nscan(height) + 1
                    scanval = scanval + image(k, maxYpos)
                end if

            end do

            scan(height) = scanval
            running_total = running_total + scanval
            nrunning_total = nrunning_total + nscan(height)
       
        
            if(type .eq. 1) then
                if (nrunning_total .gt. 0) then 
                    backgroundTot(startX, i) = running_total / nrunning_total
                else   
                    backgroundTot(startX, i) = 0.0
                end if
            else
                backgroundTot(startX, i) = running_total
            end if

        end do
    end if

    !**********************
    ! And now do end part
    !**********************
    clearEndPlus1 = clearEnd + 1  

    do i = clearEndPlus1, ny - 1
        running_total = running_total - scan(height - i + clearEndPlus1)
        nrunning_total = nrunning_total - nscan(height - i + clearEndPlus1)

        if(type .eq. 1) then
            if (nrunning_total .gt. 0) then
                backgroundTot(startX, i) = running_total / nrunning_total
            else    
                backgroundTot(startX, i) = 0.0
            end if       
        else
            backgroundTot(startX, i) = running_total
        end if
    end do

end subroutine MovingBox

!********************************************************
! subroutine smoothIMage2
! Routine to smooth image usin psfFunction 
!
!********************************************************
subroutine SmoothImage2(image, binX, binY)
 
    real(single), pointer :: image( : , :)
    integer(int32), intent(in) :: binX, binY
    integer(int32)   :: nx, ny, psfNx, psfNy
 
    real(single), pointer   :: tempImage(:, :), psfImage(:, :)
    real(double) :: xc, yc, dx1, dy1
    real(double) :: sum1, sum2, f
    type(TwoDSpline) :: spline
    
    ! internal variables
   
    integer(int32)                :: i, j, k, l, n, m
    integer(int32) :: dx, dy
  
  

    nx = size(image, 1) - 1
    ny = size(image, 2) - 1

    
    tempImage => Allocate2DSingle(0, nx, 0 , ny, "tempImage")
  
    psfNx = 25 ! Size of psf image in unbinned pixels
    psfNy = 25
    
    call CreatePSF2(1024.5_double, 1024.5_double, 1.0_double, psfNx, psfNy, binX, binY, &
               1, 1, psfSpline=spline)   

    psfImage => allocate2DSingle(0, psfNx - 1, 0, psfNy - 1, "psfImage")
write(*, *) "Smoothimage2 line 317 interpolated value at 0, 0=", InterpolateSpline(0.0_double, 0.0_double, spline)

     
    xc = real(psfNx) / 2.0
    yc = real(psfNy) / 2.0

    do i = 0, psfNx - 1
        dx1 = real(i) + 0.5_double - xc
        do j = 0, psfNy - 1
            dy1 = real(j) + 0.5_double - yc
            psfImage(i, j) = InterpolateSpline(dx1, dy1, spline)
                                    
        end do
    end do


    dx = (psfNx - 1) / 2
    dy = (psfNy - 1) / 2
    do i = 0, nx
       
        do j = 0, ny  
           
            sum1 = 0.0_double
            sum2 = 0.0_double
            n = 0 
            do k = i - dx, i + dx
           
                m = -1
                do l = j - dy, j + dy
                    m = m + 1 
                    if (l < 0 .or. l > ny .or. k < 0 .or. k > nx) cycle
                    f = psfImage(n, m)
                    sum1 = sum1 + f * image(k, l)
                    sum2 = sum2 + f
                   
                end do
                n = n + 1 
            end do
            if (sum2 > 0.0) then
                tempImage(i, j) = sum1 / sum2
            else
                tempImage(i, j) = image(i, j)
            end if     
   
        end do
    end do


    call DestructSpline(spline)    
    image = tempImage
    deallocate(tempImage)
    nullify(tempImage)
    deallocate(psfImage)
    nullify(psfImage)

end subroutine SmoothImage2

!********************************************************
! subroutine smoothIMage2
! Routine to smooth image usin psfFunction 
!
!********************************************************
subroutine SmoothImage4(image, sigma)
 
    real(single), pointer :: image( : , :)
    real(single), intent(in) :: sigma

    integer(int32)   :: nx, ny
 
    real(single), pointer   :: tempImage(:, :) => null()
    real(double) :: radius
    real(double) :: sum1, sum2, f, dx, dy
    real(double) :: constant
    
    ! internal variables
   
    integer(int32)                :: i, j, k, l
    integer(int32) :: width
  
  

    nx = ubound(image, 1)
    ny = ubound(image, 2)


    constant = -0.5_double / (sigma * sigma)

    
    tempImage => Allocate2DSingle(0, nx, 0 , ny, "tempImage")
    width = ceiling(3.0 * sigma)
 
    do i = 0, nx
       
        do j = 0, ny  
           
            sum1 = 0.0_double
            sum2 = 0.0_double
             
            do k = max(0, i - width), min(nx, i + width)
                dx = dble(k - i)
                do l = max(0, j - width), min(ny, j + width)
                    dy = dble(l - j) 
                    radius = dx * dx + dy * dy
                    f = exp(constant * radius)
                    sum1 = sum1 + f * image(k, l)
                    sum2 = sum2 + f
                   
                end do
        
            end do
            if (sum2 > 0.0) then
                tempImage(i, j) = sum1 / sum2
            else
                tempImage(i, j) = image(i, j)
            end if     
   
        end do
    end do



    image = tempImage
    deallocate(tempImage)
    nullify(tempImage)
  

end subroutine SmoothImage4



!********************************************************
! subroutine smoothIMage2
! Routine to smooth image usin psfFunction 
!
!********************************************************
subroutine PSFSmoothImage(image, binX, binY)
 
    real(single), pointer :: image( : , :)
    integer(int32), intent(in) :: binX, binY
    integer(int32)   :: nx, ny, psfNx, psfNy
 
    real(single), pointer   :: tempImage(:, :), psfImage(:, :)
    real(double) :: xc, yc, dx1, dy1
    real(double) :: sum1, sum2, f

    type(TwoDSpline) :: spline
    
    ! internal variables
   
    integer(int32)                :: i, j, k, l, n, m
    integer(int32) :: dx, dy
  
  

    nx = size(image, 1) - 1
    ny = size(image, 2) - 1

    
    tempImage => Allocate2DSingle(0, nx, 0 , ny, "tempImage")
  
    psfNx = 25 ! Size of psf image in unbinned pixels
    psfNy = 25
    
    call CreatePSF2(1024.5_double, 1024.5_double, 1.0_double, psfNx, psfNy, binX, binY, &
               1, 1, psfSpline=spline)   

    psfImage => allocate2DSingle(0, psfNx - 1, 0, psfNy - 1, "psfImage")
write(*, *) "Smoothimage2 line 317 interpolated value at 0, 0=", InterpolateSpline(0.0_double, 0.0_double, spline)

     
    xc = real(psfNx) / 2.0
    yc = real(psfNy) / 2.0

    do i = 0, psfNx - 1
        dx1 = real(i) + 0.5_double - xc
        do j = 0, psfNy - 1
            dy1 = real(j) + 0.5_double - yc
            psfImage(i, j) = InterpolateSpline(dx1, dy1, spline)
                                    
        end do
    end do


    dx = (psfNx - 1) / 2
    dy = (psfNy - 1) / 2

    do i = 0, nx
       
        do j = 0, ny  
           
            sum1 = 0.0_double
            sum2 = 0.0_double
            n = 0 
            do k = i - dx, i + dx
           
                m = -1
                do l = j - dy, j + dy
                    m = m + 1 
                    if (l < 0 .or. l > ny .or. k < 0 .or. k > nx) cycle
         
                    f = psfImage(n, m)
                    if (f .gt. 0.0) then
                        f = f * sqrt(f)
                    
                        sum1 = sum1 + f * image(k, l)
                        sum2 = sum2 + f
                    end if
                end do
                n = n + 1 
            end do
            if (sum2 > 0.0) then
                tempImage(i, j) = sum1 / sum2
            else
                tempImage(i, j) = image(i, j)
            end if     
   
        end do
    end do


    call DestructSpline(spline)    
    image = tempImage
    deallocate(tempImage)
    nullify(tempImage)
    deallocate(psfImage)
    nullify(psfImage)

end subroutine PsfSmoothImage







!*********************************************************
!
!********************************************************
subroutine KuwaharaFilter(image)
    real(single), pointer :: image(:, :)

    integer :: nx, ny
    integer :: i, j
  
    real(single), pointer :: tempImage(:, :) => null()
    real(single) :: data(6)
    real(single) :: sigmaMin, f, mean, sigma


    nx = ubound(image, 1)
    ny = ubound(image, 2)
    tempImage => allocate2dSingle(0, nx, 0, ny, "tempimage")

    tempImage = image


    nx = nx - 2
    ny = ny - 2

    do i = 2, nx
        do j = 2, ny
            !******************************************************
            ! Region 1
            !*****************************************************        
            data(1) = image(i, j + 1)
            data(2) = image(i, j + 2)
            data(3) = image(i - 1, j + 1)
            data(4) = image(i - 1, j + 2)          
            data(5) = image(i - 2, j + 1)
            data(6) = image(i - 2, j + 2)
          

          
            call ComputeStatistics1(6, data, f, sigmamin)
            !******************************************************
            ! Region 2
            !*****************************************************        
            data(1) = image(i + 1, j)
            data(2) = image(i + 1, j + 1)
            data(3) = image(i + 1, j + 2)
            data(4) = image(i + 2, j)
            data(5) = image(i + 2, j + 1)
            data(6) = image(i + 2, j + 2)
            call ComputeStatistics1(6, data, mean, sigma)
            if (sigma .lt. sigmaMin) then
                sigmaMin = sigma
                f = mean
            end if
        
            !******************************************************
            ! Region 3
            !*****************************************************        
            data(1) = image(i - 1, j)
            data(2) = image(i - 1, j - 1)
            data(3) = image(i - 1, j - 2)       
            data(4) = image(i - 2, j)
            data(5) = image(i - 2, j - 1)
            data(6) = image(i - 2, j - 2)
            call ComputeStatistics1(6, data, mean, sigma)
            if (sigma .lt. sigmaMin) then
                sigmaMin = sigma
                f = mean
            end if
        
   
            !******************************************************
            ! Region 4
            !*****************************************************        
            data(1) = image(i, j - 1)
            data(2) = image(i, j - 2)
            data(3) = image(i + 1, j - 1)
            data(4) = image(i + 1, j - 2)      
            data(5) = image(i + 2, j - 1)
            data(6) = image(i + 2, j - 2)
            call ComputeStatistics1(6, data, mean, sigma)
            if (sigma .lt. sigmaMin) then
                sigmaMin = sigma
                f = mean
            end if
        

            tempimage(i, j) = f 


        end do
    end do

    image = tempimage

    deallocate(tempimage)
    nullify(tempimage)

   


end subroutine KuwaharaFilter

end module OmSmoothImage_mod









